"""This module provides functions which determine various observatory
specific policies for HST:

1. How to convert reference file basenames to fully specified paths.

2. How to convert a reference or mapping basename to a download URL.

3. How to manage parameters for reference file Validator objects used
in the certification of reference files.

The intentions is that a similar module with different policies could
be implemented for JWST.
"""
import os.path
import datetime
import time
import warnings

# =======================================================================

from crds.core import log, rmap, config, utils, timestamp
from crds import data_file
from crds.core.exceptions import CrdsError, CrdsNamingError
from crds.hst import siname
from crds.io import abstract

# =======================================================================

# This block imports functions and globals defined in other modules to become
# part of the locator interface.

from crds.hst import TYPES, INSTRUMENTS, FILEKINDS, EXTENSIONS, INSTRUMENT_FIXERS, TYPE_FIXERS

get_row_keys_by_instrument = TYPES.get_row_keys_by_instrument
get_item = TYPES.get_item
suffix_to_filekind = TYPES.suffix_to_filekind
filekind_to_suffix = TYPES.filekind_to_suffix
get_all_tpninfos = TYPES.get_all_tpninfos

# =======================================================================
HERE = os.path.dirname(__file__) or "."

def tpn_path(tpn_file):
    return os.path.join(HERE, "tpns", tpn_file)

def get_extra_tpninfos(refpath):
    return []

def project_check(refpath, rmap):
    pass

# =======================================================================

# When loading headers,  make sure each keyword in a tuple is represented with
# the same value enabling any form to be used.
CROSS_STRAPPED_KEYWORDS = {
       # "META.INSTRUMENT.NAME" : ["INSTRUME", "INSTRUMENT", "META.INSTRUMENT.TYPE",],
    }

@utils.cached
def get_static_pairs():
    return abstract.equivalence_dict_to_pairs(CROSS_STRAPPED_KEYWORDS)

def get_cross_strapped_pairs(header):
    """Return the list of keyword pairs where each pair describes synonyms for the same
    piece of data.
    """
    return  get_static_pairs()

# =======================================================================

def header_to_reftypes(header, context="hst-operational"):
    """Based on `header` return the default list of appropriate reference type names."""
    return []  # translates to everything.

def header_to_pipelines(header, context="hst-operational"):
    """Based on `header` return the default list of appropriate reference type names."""
    raise NotImplementedError("HST has not defined header_to_pipelines().")

def get_exptypes(instrument=None):
    """Return the list of EXP_TYPE values for instrument,  or for all
    instruments if instrument is not specified.
    """
    raise NotImplementedError("HST has not defined get_exptypes().")


# =======================================================================

def match_context_key(key):
    """Set the case of a context key (instrument or type) appropriately
    for this project, HST used upper case for instruments,  lower case
    for type names.
    """
    if key.lower() in INSTRUMENTS:
        return key.upper()
    elif key.lower() in FILEKINDS:
        return key.lower()
    else:
        return None

# =======================================================================

def reference_keys_to_dataset_keys(rmapping, header):
    """Given a header dictionary for a reference file,  map the header back to
    keys relevant to datasets.
    """
    result = dict(header)

    #  XXXXX TODO   Add/consolidate logic to handle P_ pattern keywords

    # If USEAFTER is defined,  or we're configured to fake it...
    #   don't invent one if its missing and we're not faking it.
    if "USEAFTER" in header or config.ALLOW_BAD_USEAFTER:

        # Identify reference involved as best as possible
        filename = header.get("FILENAME", None) or rmapping.filename

        reformatted = timestamp.reformat_useafter(filename, header).split()
        result["DATE-OBS"] = reformatted[0]
        result["DATE_OBS"] = reformatted[0]
        result["TIME-OBS"] = reformatted[1]
        result["TIME_OBS"] = reformatted[1]

    return result

# =======================================================================

def condition_matching_header(rmapping, header):
    """Condition the matching header values to the normalized form of the .rmap"""
    return utils.condition_header(header)

# =======================================================================

def get_file_properties(filename):
    """Figure out (instrument, filekind) based on `filename` which
    should be a mapping or FITS reference file.

    >>> get_file_properties("./hst_acs_biasfile_0001.rmap")
    ('acs', 'biasfile')

    >> get_file_properties("./hst_acs_biasfile_0001.pmap")
    Traceback (most recent call last):
    ...
    IOError: [Errno 2] No such file or directory: './hst_acs_biasfile_0001.pmap'

    >> get_file_properties("test_data/s7g1700gl_dead.fits")
    """
    if data_file.is_geis_data(filename):
        # determine GEIS data file properties from corresponding header file.
        filename = filename[:-1] + "h"
    if config.is_mapping(filename):
        try:
            return decompose_newstyle_name(filename)[2:4]
        except Exception:
            return properties_inside_mapping(filename)
    elif config.is_reference(filename):
        result = get_reference_properties(filename)[2:4]
    else:
        try:
            result = properties_inside_mapping(filename)
        except Exception:
            result = get_reference_properties(filename)[2:4]
    assert result[0] in INSTRUMENTS+[""], "Bad instrument " + \
        repr(result[0]) + " in filename " + repr(filename)
    assert result[1] in FILEKINDS+[""], "Bad filekind " + \
        repr(result[1]) + " in filename " + repr(filename)
    return result

def decompose_newstyle_name(filename):
    """
    >>> decompose_newstyle_name("./hst.pmap")
    ('.', 'hst', '', '', '', '.pmap')

    >>> decompose_newstyle_name("./hst_0001.pmap")
    ('.', 'hst', '', '', '0001', '.pmap')

    >>> decompose_newstyle_name("./hst_acs.imap")
    ('.', 'hst', 'acs', '', '', '.imap')

    >>> decompose_newstyle_name("./hst_acs_0001.imap")
    ('.', 'hst', 'acs', '', '0001', '.imap')

    >>> decompose_newstyle_name("./hst_acs_biasfile.rmap")
    ('.', 'hst', 'acs', 'biasfile', '', '.rmap')

    >>> decompose_newstyle_name("./hst_acs_biasfile_0001.rmap")
    ('.', 'hst', 'acs', 'biasfile', '0001', '.rmap')

    >>> decompose_newstyle_name("./hst_acs_biasfile.fits")
    ('.', 'hst', 'acs', 'biasfile', '', '.fits')

    >>> decompose_newstyle_name("./hst_acs_biasfile_0001.fits")
    ('.', 'hst', 'acs', 'biasfile', '0001', '.fits')
    """
    path, parts, ext = _get_fields(filename)
    observatory = parts[0]
    serial = list_get(parts, 3, "")

    if ext == ".pmap":
        assert len(parts) in [1, 2], "Invalid .pmap filename " + repr(filename)
        instrument, filekind = "", ""
        serial = list_get(parts, 1, "")
    elif ext == ".imap":
        assert len(parts) in [2, 3], "Invalid .imap filename " + repr(filename)
        instrument = parts[1]
        filekind = ""
        serial = list_get(parts, 2, "")
    else:
        assert len(parts) in [3, 4], "Invalid filename " + repr(filename)
        instrument = parts[1]
        filekind = parts[2]
        serial = list_get(parts, 3, "")

    assert observatory == "hst"
    assert instrument in INSTRUMENTS+[""], "Invalid instrument " + repr(instrument)
    assert filekind in FILEKINDS+[""], "Invalid filekind " + repr(filekind)
    # assert re.fullmatch(r"\d*", serial), "Invalid id field " + repr(id)

    # extension may vary for upload temporary files.

    return path, observatory, instrument, filekind, serial, ext

def properties_inside_mapping(filename):
    """Load `filename`s mapping header to discover and return (instrument, filekind).

    >>> from crds.tests import test_config
    >>> old_config = test_config.setup()

    >>> properties_inside_mapping("hst.pmap")
    ('', '')

    >>> properties_inside_mapping("hst_acs.imap")
    ('acs', '')

    >>> properties_inside_mapping("hst_acs_darkfile.rmap")
    ('acs', 'darkfile')

    >>> test_config.cleanup(old_config)
    """
    loaded = rmap.fetch_mapping(filename)
    if loaded.mapping == "pipeline":
        result = "", ""
    elif loaded.mapping == "instrument":
        result = loaded.instrument, ""
    else:
        result = loaded.instrument, loaded.filekind
    return result

def _get_fields(filename):
    """Break CRDS-style filename down into: path, underscore-separated-parts, extension."""
    path = os.path.dirname(filename)
    name = os.path.basename(filename)
    name, ext = os.path.splitext(name)
    parts = name.split("_")
    return path, parts, ext

def list_get(the_list, index, default):
    """Fetch the `index` item from `the_list`, or return `default` on IndexError.  Like dict.get()."""
    try:
        return the_list[index]
    except IndexError:
        return default

CDBS_DIRS_TO_INSTR = {
   "/jref/":"acs",
   "/oref/":"stis",
   "/iref/":"wfc3",
   "/lref/":"cos",
   "/nref/":"nicmos",

   "/upsf/":"wfpc2",
   "/uref/":"wfpc2",
   "/uref_linux/":"wfpc2",

   "/yref/" : "fos",
   "/zref/" : "hrs",
}

def check_naming_consistency(checked_instrument=None, exhaustive_mapping_check=False):
    """Dev function to compare the properties returned by name decomposition
    to the properties determined by file contents and make sure they're the same.
    Also checks rmap membership.

    >> from crds.tests import test_config
    >> old_config = test_config.setup()
    >> check_naming_consistency("acs")
    >> check_naming_consistency("cos")
    >> check_naming_consistency("nicmos")
    >> check_naming_consistency("stis")
    >> check_naming_consistency("wfc3")
    >> check_naming_consistency("wfpc2")
    >> test_config.cleanup(old_config)
    """
    from crds import certify

    for ref in rmap.list_references("*", observatory="hst", full_path=True):
        with log.error_on_exception("Failed processing:", repr(ref)):

            _path, _observ, instrument, filekind, _serial, _ext = ref_properties_from_cdbs_path(ref)

            if checked_instrument is not None and instrument != checked_instrument:
                continue

            if data_file.is_geis_data(ref):
                if os.path.exists(data_file.get_conjugate(ref)):
                    continue
                else:
                    log.warning("No GEIS header for", repr(ref))

            log.verbose("Processing:", instrument, filekind, ref)

            _path2, _observ2, instrument2, filekind2, _serial2, _ext2 = ref_properties_from_header(ref)
            if instrument != instrument2:
                log.error("Inconsistent instruments", repr(instrument), "vs.", repr(instrument2),
                          "for", repr(ref))
            if filekind != filekind2:
                log.error("Inconsistent filekinds", repr(filekind), "vs.", repr(filekind2),
                          "for", repr(ref))

            for pmap_name in reversed(sorted(rmap.list_mappings("*.pmap", observatory="hst"))):

                r = certify.certify.find_governing_rmap(pmap_name, ref)

                if not r:
                    continue

                if r.instrument != instrument:
                    log.error("Rmap instrument", repr(r.instrument),
                              "inconsistent with name derived instrument", repr(instrument), "for", repr(ref), "in", repr(pmap_name))
                if r.filekind != filekind:
                    log.error("Rmap filekind", repr(r.filekind),
                              "inconsistent with name derived filekind", repr(filekind), "for", repr(ref), "in", repr(pmap_name))
                if r.instrument != instrument2:
                    log.error("Rmap instrument", repr(r.instrument),
                              "inconsistent with content derived instrument", repr(instrument2), "for", repr(ref), "in", repr(pmap_name))
                if r.filekind != filekind2:
                    log.error("Rmap filekind", repr(r.filekind),
                              "inconsistent with content derived filekind", repr(filekind2), "for", repr(ref), "in", repr(pmap_name))

                if not exhaustive_mapping_check:
                    break

            else:
                log.error("Orphan reference", repr(ref), "not found under any context.")

def get_reference_properties(filename):
    """Figure out FITS (instrument, filekind, serial) based on `filename`."""
    # try:
    #     return decompose_synphot_name(filename)
    # except AssertionError:
    #    pass
    try:   # Hopefully it's a nice new standard filename, easy
        return decompose_newstyle_name(filename)
    except AssertionError:  # cryptic legacy paths & names, i.e. reality
        pass
    try:   # or maybe a recognizable HST legacy path/filename, fast
        return ref_properties_from_cdbs_path(filename)
    except Exception:
        pass
    # If not, dig inside the FITS file, slow
    return ref_properties_from_header(filename)

GEIS_EXT_TO_SUFFIX = {
    "r0" : "msk",     # Static mask
    "r1" : "a2d",     # A-to-D lookup tables
    "r2" : "bas",     # Bias
    "r3" : "drk",     # Preflash
    "r4" : "flt",     # Superpurge
    "r5" : "shd",     # Shutter Shading Reference
    "r6" : "flt",     # Flat field
    "r7" : "w4t",     # wf4tfile
}

def ref_properties_from_cdbs_path(filename):
    """Based on a HST CDBS `filename`,  return

    (path, "hst", instrument, filekind, serial, ext)

    Raise AssertionError if it's not a good filename.
    """
    path, fields, extension = _get_fields(filename)
    # For legacy files,  just use the root filename as the unique id
    serial = os.path.basename(os.path.splitext(filename)[0])
    # First try to figure everything out by decoding filename. fast
    instrument = siname.WhichCDBSInstrument(os.path.basename(filename)).lower()
    if instrument == "synphot":
        if filename.endswith("_syn.fits"):
            filekind = "thruput"
        elif filename.endswith("_th.fits"):
            filekind = "thermal"
        elif filename.endswith("_tmt.fits"):
            filekind = "tmttab"
        elif filename.endswith("_tmg.fits"):
            filekind = "tmgtab"
        elif filename.endswith("_tmc.fits"):
            filekind = "tmctab"
        else:
            assert False, "Uknown synphot filetype for: " + repr(filename)
        suffix = filename.split("_")[-1][:-len(".fits")]
    elif extension == ".fits":
        suffix = fields[1]
    else:
        suffix = GEIS_EXT_TO_SUFFIX[extension[1:3]]
    try:
        filekind = TYPES.suffix_to_filekind(instrument, suffix)
    except KeyError:
        assert False, "Couldn't map extension/suffix " + repr(suffix) + " to filekind."
    return path, "hst", instrument, filekind, serial, extension

def instrument_from_refname(filename):
    """Based on `filename` rather than it's contents,  determine the associated
    instrument or raise an exception.

    >>> instrument_from_refname('hst_cos_spottab_0052.fits')
    'cos'

    >>> instrument_from_refname('zas1615jl_spot.fits')
    'cos'

    >>> instrument_from_refname('foobar.fits')
    Traceback (most recent call last):
    ...
    AssertionError: Cannot determine instrument for filename 'foobar.fits'
    """
    try:   # Hopefully it's a nice new standard filename, easy
        return decompose_newstyle_name(filename)[2]
    except AssertionError:  # cryptic legacy paths & names, i.e. reality
        try:
            instrument = siname.WhichCDBSInstrument(os.path.basename(filename)).lower()
            if instrument == "multi":
                instrument = "synphot"
            return instrument
        except Exception:
            assert False, "Cannot determine instrument for filename '{}'".format(filename)

def ref_properties_from_header(filename):
    """Look inside FITS `filename` header to determine:

    (path, "hst", instrument, filekind, serial, ext)
    """
    # For legacy files,  just use the root filename as the unique id
    path, _parts, ext = _get_fields(filename)
    serial = os.path.basename(os.path.splitext(filename)[0])
    header = data_file.get_free_header(filename, (), None, "hst")
    try:
        if "DBTABLE" not in header or header["DBTABLE"] in ["IMPHTTAB"]:
            instrument = header["INSTRUME"].lower()
        else:
            instrument = "synphot"
        instrument = INSTRUMENT_FIXERS.get(instrument, instrument)
    except KeyError as exc:
        raise CrdsNamingError("Can't determine instrument for",
                              repr(os.path.basename(filename)) + ".",
                              "CAL references must define INSTRUME,",
                              "SYNPHOT references define DBTABLE.") from exc
    try:
        filetype = header.get(
            "FILETYPE", header.get(
                "DBTABLE", header.get("CDBSFILE")))
        if filetype is None:
            raise KeyError
        else:
            filetype = filetype.lower()
    except KeyError:
        observatory = header.get(
            "TELESCOP", header.get(
                "TELESCOPE", None))
        if observatory is not None and observatory.upper() != "HST":
            raise CrdsNamingError(
                "CRDS is configured for 'HST' but file",
                repr(os.path.basename(filename)),
                "is for the", repr(observatory),
                "telescope.  Reconfigure CRDS_PATH or CRDS_SEVER_URL.")
        else:
            raise CrdsNamingError(
                "Can't determine HST file type for", repr(os.path.basename(filename)) + ".",
                "Check FILETYPE, CDBSFILE, DBTABLE, TELESCOP, and/or TELESCOPE.")
    filetype = TYPE_FIXERS.get((instrument, filetype), filetype)
    try:
        filekind = TYPES.filetype_to_filekind(instrument, filetype)
    except KeyError:
        raise CrdsNamingError(f"Invalid FILETYPE (or CDBSFILE) of '{filetype}' for instrument '{instrument}'.")
    return path, "hst", instrument, filekind, serial, ext

# ============================================================================

def generate_unique_name(filename, now=None):

    """Given an arbitrarily named filename (which must correctly define it's format, e.g. .fits)
    generate and return a unique enhanced CDBS-style name which incorporates a timestamp,
    and instrument id character, and a filetype suffix.

Character 1   : Year [z=2015, 0=2016, 1= 2017, etc.]
Character 2   : Month [1-9, a-c]
Character 3   : Day [1-9, a-s(28), t(29), u(30), v(31)]
Character 4-5: UT Hour [00 - 23]
Character 6-7: UT Minute [00 - 59]
Character 8   : UT Seconds [0-9, a-t (~2 second intervals)]
Character 9   : Instrument Designation [j=ACS, i=WFC3, o=STIS, l=COS,
u=WFPC2, n=NICMOS, m=SYNPHOT]
    """
    path, _obs, instr, filekind, _serial, ext = get_reference_properties(filename)

    name = generate_unique_name_core(instr, filekind, ext, now)

    return os.path.join(path, name)

def generate_unique_name_core(instr, filekind, extension, now=None):
    """Given an arbitrarily named filename (which must correctly define it's format, e.g. .fits)
    generate and return a unique enhanced CDBS-style name which incorporates a timestamp,
    and instrument id character, and a filetype suffix.

Character 1   : Year [z=2015, 0=2016, 1= 2017, etc.]
Character 2   : Month [1-9, a-c]
Character 3   : Day [1-9, a-s(28), t(29), u(30), v(31)]
Character 4-5: UT Hour [00 - 23]
Character 6-7: UT Minute [00 - 59]
Character 8   : UT Seconds [0-9, a-t (~2 second intervals)]
Character 9   : Instrument Designation [j=ACS, i=WFC3, o=STIS, l=COS,
u=WFPC2, n=NICMOS, m=MULTI, m=SYNPHOT]
    """
    if now is None:   # delay to ensure timestamp is unique.
        time.sleep(2)

    timeid = generate_timestamp(now)

    suffix = "_" + filekind_to_suffix(instr, filekind)
    instr_char = siname.instrument_to_id_char(instr)

    return timeid + instr_char + suffix + extension

def generate_timestamp(now=None):

    """Generate an enhanced CDBS-style uniqname."""

    if now is None:
        now = datetime.datetime.utcnow()

    if now.year < 2016:
        year = chr(now.year - 2015 + ord('z'))
    elif 2016 <= now.year <= 2025:
        year = chr(now.year - 2016 + ord('0'))
    else:
        raise RuntimeError("Unique names are not defined for 2026 and beyond.")

    if 1 <= now.month <= 9:
        month = chr(now.month - 1 + ord('1'))
    elif 10 <= now.month <= 12:
        month = chr(now.month - 10 + ord('a'))

    if 1 <= now.day <= 9:
        day = chr(now.day - 1 + ord('1'))
    elif 10 <= now.day <= 31:
        day = chr(now.day - 10 + ord('a'))

    hour = "%02d" % now.hour
    minute = "%02d" % now.minute

    hsecs = now.second // 2
    if 0 <= hsecs <= 9:
        second = chr(hsecs + ord('0'))
    elif 10 <= hsecs <= 59:
        second = chr(hsecs - 10 + ord('a'))

    return "".join([year, month, day, hour, minute, second])

# ============================================================================

# HST FITS headers have filenames adorned with environment prefixes for each
# instrument.

def get_env_prefix(instrument):
    """Return the environment variable prefix (IRAF prefix) for `instrument`."""
    return siname.add_IRAF_prefix(instrument.upper())

def filekind_to_keyword(filekind):
    """Return the FITS keyword at which a reference should be recorded."""
    return filekind.upper()

def locate_file(refname, mode=None):
    """Given a valid reffilename in CDBS or CRDS format,  return a cache path for the file.
    The aspect of this which is complicated is determining instrument and an instrument
    specific sub-directory for it based on the filename alone,  not the file contents.
    """
    try:
        instrument = instrument_from_refname(refname)
    except Exception:
        instrument = get_reference_properties(refname)[1]
    rootdir = locate_dir(instrument, mode)
    return  os.path.join(rootdir, os.path.basename(refname))

def locate_dir(instrument, mode=None):
    """Locate the instrument specific directory for a reference file."""
    if mode is  None:
        mode = config.get_crds_ref_subdir_mode(observatory="hst")
    else:
        config.check_crds_ref_subdir_mode(mode)
    crds_refpath = config.get_crds_refpath("hst")
    prefix = get_env_prefix(instrument)
    if mode == "legacy":   # Locate cached files at the appropriate CDBS-style  iref$ locations
        try:
            rootdir = os.environ[prefix]
        except KeyError:
            try:
                rootdir = os.environ[prefix[:-1]]
            except KeyError:
                raise KeyError("Reference location not defined for " + repr(instrument) +
                               ".  Did you configure " + repr(prefix) + "?")
    elif mode == "instrument" and instrument != "synphot":   # use simple names inside CRDS cache.
        rootdir = os.path.join(crds_refpath, instrument)
        refdir = os.path.join(crds_refpath, prefix[:-1])
        if not os.path.exists(refdir):
            if config.writable_cache_or_verbose("Skipping making instrument directory link for", repr(instrument)):
                log.verbose("Creating legacy cache link", repr(refdir), "-->", repr(rootdir))
                with log.verbose_warning_on_exception("Failed creating legacy symlink:", refdir, "-->", rootdir):
                    utils.ensure_dir_exists(rootdir + "/locate_dir.fits")
                    os.symlink(rootdir, refdir)
    elif mode == "instrument" and instrument == "synphot":
        rootdir = os.path.join(crds_refpath, instrument)
    elif mode == "flat":    # use original flat cache structure,  all instruments in same directory.
        rootdir = crds_refpath
    else:
        raise ValueError("Unhandled reference file location mode " + repr(mode))
    return rootdir

# ============================================================================

def fits_to_parkeys(header):
    """Map a FITS header onto rmap parkeys appropriate for this observatory."""
    return dict(header)

# =======================================================================

def test():
    """Run the module doctests."""
    import doctest
    from crds.hst import locate
    return doctest.testmod(locate)

# =======================================================================

if __name__ == "__main__":
    print(test())



"""This module tests the synphot pseudo-instrument support for HST."""

# ==============================================================
import crds
from crds.core import config, log, naming, timestamp
from crds.hst import locate as hst_locate
from crds import certify
from crds import diff
from crds.refactoring import refactor

from crds.tests import test_config

# ==============================================================

def dt_synphot_naming():
    """
    >>> old_state = test_config.setup()

    >>> NOW = timestamp.parse_date("2018-11-14T00:00:00")

    THERMAL

    >>> config.is_crds_name("wfc3_ir_f098m_002_th.fits")
    False
    >>> config.is_cdbs_name("wfc3_ir_f098m_002_th.fits")
    True
    >>> config.is_reference("wfc3_ir_f098m_002_th.fits")
    True
    >>> config.is_mapping("wfc3_ir_f098m_002_th.fits")
    False
    >>> hst_locate.get_file_properties("wfc3_ir_f098m_002_th.fits")
    ('synphot', 'thermal')
    >>> hst_locate.ref_properties_from_cdbs_path("data/wfc3_ir_f098m_002_th.fits")
    ('data', 'hst', 'synphot', 'thermal', 'wfc3_ir_f098m_002_th', '.fits')
    >>> hst_locate.ref_properties_from_header("data/wfc3_ir_f098m_002_th.fits")
    ('data', 'hst', 'synphot', 'thermal', 'wfc3_ir_f098m_002_th', '.fits')
    >>> naming.generate_unique_name("wfc3_ir_f098m_002_th.fits", "hst", NOW)
    '2be00000m_th.fits'

    THRUPUT

    >>> config.is_crds_name("wfc3_uvis_f469nf2_003_syn.fits")
    False
    >>> config.is_cdbs_name("wfc3_uvis_f469nf2_003_syn.fits")
    True
    >>> config.is_reference("wfc3_uvis_f469nf2_003_syn.fits")
    True
    >>> config.is_mapping("wfc3_uvis_f469nf2_003_syn.fits")
    False
    >>> hst_locate.get_file_properties("wfc3_uvis_f469nf2_003_syn.fits")
    ('synphot', 'thruput')
    >>> hst_locate.ref_properties_from_cdbs_path("data/wfc3_uvis_f469nf2_003_syn.fits")
    ('data', 'hst', 'synphot', 'thruput', 'wfc3_uvis_f469nf2_003_syn', '.fits')
    >>> hst_locate.ref_properties_from_header("data/wfc3_uvis_f469nf2_003_syn.fits")
    ('data', 'hst', 'synphot', 'thruput', 'wfc3_uvis_f469nf2_003_syn', '.fits')
    >>> naming.generate_unique_name("wfc3_uvis_f469nf2_003_syn.fits", "hst", NOW)
    '2be00000m_syn.fits'

    TMGTAB

    >>> config.is_crds_name("2381905mm_tmg.fits")
    False
    >>> config.is_cdbs_name("2381905mm_tmg.fits")
    True
    >>> config.is_reference("2381905mm_tmg.fits")
    True
    >>> config.is_mapping("2381905mm_tmg.fits")
    False
    >>> hst_locate.get_file_properties("2381905mm_tmg.fits")
    ('synphot', 'tmgtab')
    >>> hst_locate.ref_properties_from_cdbs_path("data/2381905mm_tmg.fits")
    ('data', 'hst', 'synphot', 'tmgtab', '2381905mm_tmg', '.fits')
    >>> hst_locate.ref_properties_from_header("data/2381905mm_tmg.fits")
    ('data', 'hst', 'synphot', 'tmgtab', '2381905mm_tmg', '.fits')
    >>> naming.generate_unique_name("2381905mm_tmg.fits", "hst", NOW)
    '2be00000m_tmg.fits'

    TMCTAB

    >>> config.is_crds_name("43h1909cm_tmc.fits")
    False
    >>> config.is_cdbs_name("43h1909cm_tmc.fits")
    True
    >>> config.is_reference("43h1909cm_tmc.fits")
    True
    >>> config.is_mapping("43h1909cm_tmc.fits")
    False
    >>> hst_locate.get_file_properties("43h1909cm_tmc.fits")
    ('synphot', 'tmctab')
    >>> hst_locate.ref_properties_from_cdbs_path("data/43h1909cm_tmc.fits")
    ('data', 'hst', 'synphot', 'tmctab', '43h1909cm_tmc', '.fits')
    >>> hst_locate.ref_properties_from_header("data/43h1909cm_tmc.fits")
    ('data', 'hst', 'synphot', 'tmctab', '43h1909cm_tmc', '.fits')
    >>> naming.generate_unique_name("43h1909cm_tmc.fits", "hst", NOW)
    '2be00000m_tmc.fits'

    TMTTAB

    >>> config.is_crds_name("tae17277m_tmt.fits")
    False
    >>> config.is_cdbs_name("tae17277m_tmt.fits")
    True
    >>> config.is_reference("tae17277m_tmt.fits")
    True
    >>> config.is_mapping("tae17277m_tmt.fits")
    False
    >>> hst_locate.get_file_properties("tae17277m_tmt.fits")
    ('synphot', 'tmttab')
    >>> hst_locate.ref_properties_from_header("data/tae17277m_tmt.fits")
    ('data', 'hst', 'synphot', 'tmttab', 'tae17277m_tmt', '.fits')
    >>> hst_locate.ref_properties_from_cdbs_path("data/tae17277m_tmt.fits")
    ('data', 'hst', 'synphot', 'tmttab', 'tae17277m_tmt', '.fits')
    >>> naming.generate_unique_name("tae17277m_tmt.fits", "hst", NOW)
    '2be00000m_tmt.fits'

    >>> test_config.cleanup(old_state)
    """

def dt_synphot_certify_refs():
    """
    >>> old_state = test_config.setup()

    TMC   reference

    >>> certify.CertifyScript("crds.certify  --run-fitsverify data/43h1909cm_tmc.fits --comparison-context hst_0772.pmap")() # doctest: +ELLIPSIS
    CRDS - INFO -  ########################################
    CRDS - INFO -  Certifying 'data/43h1909cm_tmc.fits' (1/1) as 'FITS' relative to context 'hst_0772.pmap'
    CRDS - INFO -  FITS file '43h1909cm_tmc.fits' conforms to FITS standards.
    CRDS - INFO -  Comparing reference '43h1909cm_tmc.fits' against '43h1240om_tmc.fits'
    CRDS - INFO -  Mode columns defined by spec for old reference '43h1240om_tmc.fits[1]' are: ['COMPNAME']
    CRDS - INFO -  All column names for this table old reference '43h1240om_tmc.fits[1]' are: ['TIME', 'COMPNAME', 'FILENAME', 'COMMENT']
    CRDS - INFO -  Checking for duplicate modes using intersection ['COMPNAME']
    CRDS - INFO -  Mode columns defined by spec for new reference '43h1909cm_tmc.fits[1]' are: ['COMPNAME']
    CRDS - INFO -  All column names for this table new reference '43h1909cm_tmc.fits[1]' are: ['TIME', 'COMPNAME', 'FILENAME', 'COMMENT']
    CRDS - INFO -  Checking for duplicate modes using intersection ['COMPNAME']
    CRDS - INFO -  Running fitsverify.
    CRDS - INFO -  >>
    CRDS - INFO -  >>               fitsverify ... (CFITSIO ...)
    CRDS - INFO -  >>               --------------------------------
    CRDS - INFO -  >>
    CRDS - INFO -  >>
    CRDS - INFO -  >> File: data/43h1909cm_tmc.fits
    CRDS - INFO -  >>
    CRDS - INFO -  >> 2 Header-Data Units in this file.
    CRDS - INFO -  >>
    CRDS - INFO -  >> =================== HDU 1: Primary Array ===================
    CRDS - INFO -  >>
    CRDS - INFO -  >>  121 header keywords
    CRDS - INFO -  >>
    CRDS - INFO -  >>  Null data array; NAXIS = 0
    CRDS - INFO -  >>
    CRDS - INFO -  >> =================== HDU 2: BINARY Table ====================
    CRDS - INFO -  >>
    CRDS - INFO -  >>  21 header keywords
    CRDS - INFO -  >>
    CRDS - INFO -  >>    (4 columns x 2713 rows)
    CRDS - INFO -  >>
    CRDS - INFO -  >>  Col# Name (Units)       Format
    CRDS - INFO -  >>    1 TIME                 26A
    CRDS - INFO -  >>    2 COMPNAME             18A
    CRDS - INFO -  >>    3 FILENAME             56A
    CRDS - INFO -  >>    4 COMMENT              68A
    CRDS - INFO -  >>
    CRDS - INFO -  >> ++++++++++++++++++++++ Error Summary  ++++++++++++++++++++++
    CRDS - INFO -  >>
    CRDS - INFO -  >>  HDU#  Name (version)       Type             Warnings  Errors
    CRDS - INFO -  >>  1                          Primary Array    0         0
    CRDS - INFO -  >>  2                          Binary Table     0         0
    CRDS - INFO -  >>
    CRDS - INFO -  >> **** Verification found 0 warning(s) and 0 error(s). ****
    CRDS - INFO -  ########################################
    CRDS - INFO -  0 errors
    CRDS - INFO -  0 warnings
    CRDS - INFO -  46 infos
    0

    >>> certify.CertifyScript("crds.certify  --run-fitsverify data/2381905mm_tmg.fits --comparison-context hst_0691.pmap")()   # doctest: +ELLIPSIS
    CRDS - INFO -  ########################################
    CRDS - INFO -  Certifying 'data/2381905mm_tmg.fits' (1/1) as 'FITS' relative to context 'hst_0691.pmap'
    CRDS - INFO -  FITS file '2381905mm_tmg.fits' conforms to FITS standards.
    CRDS - INFO -  Mode columns defined by spec for new reference '2381905mm_tmg.fits[1]' are: ['COMPNAME', 'KEYWORD', 'INNODE', 'OUTNODE', 'THCOMPNAME']
    CRDS - INFO -  All column names for this table new reference '2381905mm_tmg.fits[1]' are: ['COMPNAME', 'KEYWORD', 'INNODE', 'OUTNODE', 'THCOMPNAME', 'COMMENT']
    CRDS - INFO -  Checking for duplicate modes using intersection ['COMPNAME', 'INNODE', 'KEYWORD', 'OUTNODE', 'THCOMPNAME']
    CRDS - WARNING -  No comparison reference for '2381905mm_tmg.fits' in context 'hst_0691.pmap'. Skipping tables comparison.
    CRDS - INFO -  Running fitsverify.
    CRDS - INFO -  >>
    CRDS - INFO -  >>               fitsverify ... (CFITSIO ...)
    CRDS - INFO -  >>               --------------------------------
    CRDS - INFO -  >>
    CRDS - INFO -  >>
    CRDS - INFO -  >> File: data/2381905mm_tmg.fits
    CRDS - INFO -  >>
    CRDS - INFO -  >> 2 Header-Data Units in this file.
    CRDS - INFO -  >>
    CRDS - INFO -  >> =================== HDU 1: Primary Array ===================
    CRDS - INFO -  >>
    CRDS - INFO -  >>  192 header keywords
    CRDS - INFO -  >>
    CRDS - INFO -  >>  Null data array; NAXIS = 0
    CRDS - INFO -  >>
    CRDS - INFO -  >> =================== HDU 2: BINARY Table ====================
    CRDS - INFO -  >>
    CRDS - INFO -  >>  103 header keywords
    CRDS - INFO -  >>
    CRDS - INFO -  >>    (6 columns x 3656 rows)
    CRDS - INFO -  >>
    CRDS - INFO -  >>  Col# Name (Units)       Format
    CRDS - INFO -  >>    1 COMPNAME             18A
    CRDS - INFO -  >>    2 KEYWORD              11A
    CRDS - INFO -  >>    3 INNODE               J
    CRDS - INFO -  >>    4 OUTNODE              J
    CRDS - INFO -  >>    5 THCOMPNAME           17A
    CRDS - INFO -  >>    6 COMMENT              68A
    CRDS - INFO -  >>
    CRDS - INFO -  >> ++++++++++++++++++++++ Error Summary  ++++++++++++++++++++++
    CRDS - INFO -  >>
    CRDS - INFO -  >>  HDU#  Name (version)       Type             Warnings  Errors
    CRDS - INFO -  >>  1                          Primary Array    0         0
    CRDS - INFO -  >>  2                          Binary Table     0         0
    CRDS - INFO -  >>
    CRDS - INFO -  >> **** Verification found 0 warning(s) and 0 error(s). ****
    CRDS - INFO -  ########################################
    CRDS - INFO -  0 errors
    CRDS - INFO -  1 warnings
    CRDS - INFO -  44 infos
    0

    TMT   reference

    >>> certify.CertifyScript("crds.certify  --run-fitsverify data/tae17277m_tmt.fits --comparison-context hst_0691.pmap")()   # doctest: +ELLIPSIS
    CRDS - INFO -  ########################################
    CRDS - INFO -  Certifying 'data/tae17277m_tmt.fits' (1/1) as 'FITS' relative to context 'hst_0691.pmap'
    CRDS - INFO -  FITS file 'tae17277m_tmt.fits' conforms to FITS standards.
    CRDS - ERROR -  Malformed FILENAME value at index 1 (missing or invalid path prefix)
    CRDS - INFO -  Comparing reference 'tae17277m_tmt.fits' against '3241637sm_tmt.fits'
    CRDS - INFO -  Mode columns defined by spec for old reference '3241637sm_tmt.fits[1]' are: ['COMPNAME']
    CRDS - INFO -  All column names for this table old reference '3241637sm_tmt.fits[1]' are: ['TIME', 'COMPNAME', 'FILENAME', 'COMMENT']
    CRDS - INFO -  Checking for duplicate modes using intersection ['COMPNAME']
    CRDS - INFO -  Mode columns defined by spec for new reference 'tae17277m_tmt.fits[1]' are: ['COMPNAME']
    CRDS - INFO -  All column names for this table new reference 'tae17277m_tmt.fits[1]' are: ['TIME', 'COMPNAME', 'FILENAME', 'COMMENT']
    CRDS - INFO -  Checking for duplicate modes using intersection ['COMPNAME']
    CRDS - INFO -  Table mode (('COMPNAME', 'dark'),) of new reference 'tae17277m_tmt.fits[1]' is NOT IN old reference '3241637sm_tmt.fits'
    CRDS - INFO -  Table mode (('COMPNAME', 'nic1_dn'),) of new reference 'tae17277m_tmt.fits[1]' is NOT IN old reference '3241637sm_tmt.fits'
    CRDS - INFO -  Table mode (('COMPNAME', 'nic2_dn'),) of new reference 'tae17277m_tmt.fits[1]' is NOT IN old reference '3241637sm_tmt.fits'
    CRDS - INFO -  Table mode (('COMPNAME', 'nic3_dn'),) of new reference 'tae17277m_tmt.fits[1]' is NOT IN old reference '3241637sm_tmt.fits'
    CRDS - INFO -  Running fitsverify.
    CRDS - INFO -  >>
    CRDS - INFO -  >>               fitsverify ... (CFITSIO ...)
    CRDS - INFO -  >>               --------------------------------
    CRDS - INFO -  >>
    CRDS - INFO -  >>
    CRDS - INFO -  >> File: data/tae17277m_tmt.fits
    CRDS - INFO -  >>
    CRDS - INFO -  >> 2 Header-Data Units in this file.
    CRDS - INFO -  >>
    CRDS - INFO -  >> =================== HDU 1: Primary Array ===================
    CRDS - INFO -  >>
    CRDS - INFO -  >>  32 header keywords
    CRDS - INFO -  >>
    CRDS - INFO -  >>  Null data array; NAXIS = 0
    CRDS - INFO -  >>
    CRDS - INFO -  >> =================== HDU 2: BINARY Table ====================
    CRDS - INFO -  >>
    CRDS - INFO -  >>  54 header keywords
    CRDS - INFO -  >>
    CRDS - INFO -  >>    (4 columns x 144 rows)
    CRDS - INFO -  >>
    CRDS - INFO -  >>  Col# Name (Units)       Format
    CRDS - INFO -  >>    1 TIME                 27A
    CRDS - INFO -  >>    2 COMPNAME             18A
    CRDS - INFO -  >>    3 FILENAME             50A
    CRDS - INFO -  >>    4 COMMENT              68A
    CRDS - INFO -  >>
    CRDS - INFO -  >> ++++++++++++++++++++++ Error Summary  ++++++++++++++++++++++
    CRDS - INFO -  >>
    CRDS - INFO -  >>  HDU#  Name (version)       Type             Warnings  Errors
    CRDS - INFO -  >>  1                          Primary Array    0         0
    CRDS - INFO -  >>  2                          Binary Table     0         0
    CRDS - INFO -  >>
    CRDS - INFO -  >> **** Verification found 0 warning(s) and 0 error(s). ****
    CRDS - INFO -  ########################################
    CRDS - INFO -  1 errors
    CRDS - INFO -  0 warnings
    CRDS - INFO -  50 infos
    1

    THERMAL reference

    >>> certify.CertifyScript("crds.certify  --run-fitsverify data/wfc3_ir_f098m_002_th.fits --comparison-context hst_0691.pmap")()  # doctest: +ELLIPSIS
    CRDS - INFO -  ########################################
    CRDS - INFO -  Certifying 'data/wfc3_ir_f098m_002_th.fits' (1/1) as 'FITS' relative to context 'hst_0691.pmap'
    CRDS - INFO -  FITS file 'wfc3_ir_f098m_002_th.fits' conforms to FITS standards.
    CRDS - ERROR -  New filename version (002) must exceed previous version (002)
    CRDS - INFO -  Mode columns defined by spec for new reference 'wfc3_ir_f098m_002_th.fits[1]' are: ['WAVELENGTH']
    CRDS - INFO -  All column names for this table new reference 'wfc3_ir_f098m_002_th.fits[1]' are: ['WAVELENGTH', 'EMISSIVITY']
    CRDS - INFO -  Checking for duplicate modes using intersection ['WAVELENGTH']
    CRDS - WARNING -  No comparison reference for 'wfc3_ir_f098m_002_th.fits' in context 'hst_0691.pmap'. Skipping tables comparison.
    CRDS - INFO -  Running fitsverify.
    CRDS - INFO -  >>
    CRDS - INFO -  >>               fitsverify ... (CFITSIO V...)
    CRDS - INFO -  >>               --------------------------------
    CRDS - INFO -  >>
    CRDS - INFO -  >>
    CRDS - INFO -  >> File: data/wfc3_ir_f098m_002_th.fits
    CRDS - INFO -  >>
    CRDS - INFO -  >> 2 Header-Data Units in this file.
    CRDS - INFO -  >>
    CRDS - INFO -  >> =================== HDU 1: Primary Array ===================
    CRDS - INFO -  >>
    CRDS - INFO -  >>  17 header keywords
    CRDS - INFO -  >>
    CRDS - INFO -  >>  Null data array; NAXIS = 0
    CRDS - INFO -  >>
    CRDS - INFO -  >> =================== HDU 2: BINARY Table ====================
    CRDS - INFO -  >>
    CRDS - INFO -  >>  20 header keywords
    CRDS - INFO -  >>
    CRDS - INFO -  >>    (2 columns x 5863 rows)
    CRDS - INFO -  >>
    CRDS - INFO -  >>  Col# Name (Units)       Format
    CRDS - INFO -  >>    1 WAVELENGTH (ANGSTROM E
    CRDS - INFO -  >>    2 EMISSIVITY           E
    CRDS - INFO -  >>
    CRDS - INFO -  >> ++++++++++++++++++++++ Error Summary  ++++++++++++++++++++++
    CRDS - INFO -  >>
    CRDS - INFO -  >>  HDU#  Name (version)       Type             Warnings  Errors
    CRDS - INFO -  >>  1                          Primary Array    0         0
    CRDS - INFO -  >>  2                          Binary Table     0         0
    CRDS - INFO -  >>
    CRDS - INFO -  >> **** Verification found 0 warning(s) and 0 error(s). ****
    CRDS - INFO -  ########################################
    CRDS - INFO -  1 errors
    CRDS - INFO -  1 warnings
    CRDS - INFO -  40 infos
    1

    THRUPUT reference

    >>> certify.CertifyScript("crds.certify  --run-fitsverify data/wfc3_uvis_f469nf2_003_syn.fits --comparison-context hst_0691.pmap")() # doctest: +ELLIPSIS
    CRDS - INFO -  ########################################
    CRDS - INFO -  Certifying 'data/wfc3_uvis_f469nf2_003_syn.fits' (1/1) as 'FITS' relative to context 'hst_0691.pmap'
    CRDS - INFO -  FITS file 'wfc3_uvis_f469nf2_003_syn.fits' conforms to FITS standards.
    CRDS - ERROR -  New filename version (003) must exceed previous version (003)
    CRDS - INFO -  Mode columns defined by spec for new reference 'wfc3_uvis_f469nf2_003_syn.fits[1]' are: ['WAVELENGTH']
    CRDS - INFO -  All column names for this table new reference 'wfc3_uvis_f469nf2_003_syn.fits[1]' are: ['WAVELENGTH', 'THROUGHPUT']
    CRDS - INFO -  Checking for duplicate modes using intersection ['WAVELENGTH']
    CRDS - WARNING -  No comparison reference for 'wfc3_uvis_f469nf2_003_syn.fits' in context 'hst_0691.pmap'. Skipping tables comparison.
    CRDS - INFO -  Running fitsverify.
    CRDS - INFO -  >>
    CRDS - INFO -  >>               fitsverify ... (CFITSIO V...)
    CRDS - INFO -  >>               --------------------------------
    CRDS - INFO -  >>
    CRDS - INFO -  >>
    CRDS - INFO -  >> File: data/wfc3_uvis_f469nf2_003_syn.fits
    CRDS - INFO -  >>
    CRDS - INFO -  >> 2 Header-Data Units in this file.
    CRDS - INFO -  >>
    CRDS - INFO -  >> =================== HDU 1: Primary Array ===================
    CRDS - INFO -  >>
    CRDS - INFO -  >>  24 header keywords
    CRDS - INFO -  >>
    CRDS - INFO -  >>  Null data array; NAXIS = 0
    CRDS - INFO -  >>
    CRDS - INFO -  >> =================== HDU 2: BINARY Table ====================
    CRDS - INFO -  >>
    CRDS - INFO -  >>  17 header keywords
    CRDS - INFO -  >>
    CRDS - INFO -  >>    (2 columns x 13 rows)
    CRDS - INFO -  >>
    CRDS - INFO -  >>  Col# Name (Units)       Format
    CRDS - INFO -  >>    1 WAVELENGTH (ANGSTROM E
    CRDS - INFO -  >>    2 THROUGHPUT (TRANSMIS E
    CRDS - INFO -  >>
    CRDS - INFO -  >> ++++++++++++++++++++++ Error Summary  ++++++++++++++++++++++
    CRDS - INFO -  >>
    CRDS - INFO -  >>  HDU#  Name (version)       Type             Warnings  Errors
    CRDS - INFO -  >>  1                          Primary Array    0         0
    CRDS - INFO -  >>  2                          Binary Table     0         0
    CRDS - INFO -  >>
    CRDS - INFO -  >> **** Verification found 0 warning(s) and 0 error(s). ****
    CRDS - INFO -  ########################################
    CRDS - INFO -  1 errors
    CRDS - INFO -  1 warnings
    CRDS - INFO -  40 infos
    1

    >>> test_config.cleanup(old_state)
    """

def dt_synphot_certify_rmaps():
    """
    >>> old_state = test_config.setup()

    TMC   rmap

    >>> certify.CertifyScript("crds.certify data/synphot_tmctab.rmap --comparison-context hst_0857.pmap")() # doctest: +ELLIPSIS
    CRDS - INFO -  ########################################
    CRDS - INFO -  Certifying 'data/synphot_tmctab.rmap' (1/1) as 'MAPPING' relative to context 'hst_0857.pmap'
    CRDS - INFO -  Mapping 'synphot_tmctab.rmap' corresponds to 'hst_synphot_tmctab_0021.rmap' from context 'hst_0857.pmap' for checking mapping differences.
    CRDS - INFO -  Checking diffs from 'hst_synphot_tmctab_0021.rmap' to 'synphot_tmctab.rmap'
    CRDS - INFO -  ########################################
    CRDS - INFO -  0 errors
    CRDS - INFO -  0 warnings
    CRDS - INFO -  5 infos
    0

    TMG   rmap

    >>> certify.CertifyScript("crds.certify data/synphot_tmgtab.rmap --comparison-context hst_0857.pmap")() # doctest: +ELLIPSIS
    CRDS - INFO -  ########################################
    CRDS - INFO -  Certifying 'data/synphot_tmgtab.rmap' (1/1) as 'MAPPING' relative to context 'hst_0857.pmap'
    CRDS - INFO -  Mapping 'synphot_tmgtab.rmap' corresponds to 'hst_synphot_tmgtab_0007.rmap' from context 'hst_0857.pmap' for checking mapping differences.
    CRDS - INFO -  Checking diffs from 'hst_synphot_tmgtab_0007.rmap' to 'synphot_tmgtab.rmap'
    CRDS - INFO -  ########################################
    CRDS - INFO -  0 errors
    CRDS - INFO -  0 warnings
    CRDS - INFO -  5 infos
    0

    TMT   rmap

    >>> certify.CertifyScript("crds.certify data/synphot_tmttab.rmap --comparison-context hst_0857.pmap")() # doctest: +ELLIPSIS
    CRDS - INFO -  ########################################
    CRDS - INFO -  Certifying 'data/synphot_tmttab.rmap' (1/1) as 'MAPPING' relative to context 'hst_0857.pmap'
    CRDS - INFO -  Mapping 'synphot_tmttab.rmap' corresponds to 'hst_synphot_tmttab_0001.rmap' from context 'hst_0857.pmap' for checking mapping differences.
    CRDS - INFO -  Checking diffs from 'hst_synphot_tmttab_0001.rmap' to 'synphot_tmttab.rmap'
    CRDS - INFO -  ########################################
    CRDS - INFO -  0 errors
    CRDS - INFO -  0 warnings
    CRDS - INFO -  5 infos
    0

    THERMAL rmap

    >>> certify.CertifyScript("crds.certify data/synphot_thermal.rmap --comparison-context hst_0857.pmap")() # doctest: +ELLIPSIS
    CRDS - INFO -  ########################################
    CRDS - INFO -  Certifying 'data/synphot_thermal.rmap' (1/1) as 'MAPPING' relative to context 'hst_0857.pmap'
    CRDS - INFO -  Mapping 'synphot_thermal.rmap' corresponds to 'hst_synphot_thermal_0002.rmap' from context 'hst_0857.pmap' for checking mapping differences.
    CRDS - INFO -  Checking diffs from 'hst_synphot_thermal_0002.rmap' to 'synphot_thermal.rmap'
    CRDS - INFO -  ########################################
    CRDS - INFO -  0 errors
    CRDS - INFO -  0 warnings
    CRDS - INFO -  5 infos
    0

    THRUPUT rmap
    >>> certify.CertifyScript("crds.certify data/synphot_thruput.rmap --comparison-context hst_0857.pmap")() # doctest: +ELLIPSIS
    CRDS - INFO -  ########################################
    CRDS - INFO -  Certifying 'data/synphot_thruput.rmap' (1/1) as 'MAPPING' relative to context 'hst_0857.pmap'
    CRDS - INFO -  Mapping 'synphot_thruput.rmap' corresponds to 'hst_synphot_thruput_0025.rmap' from context 'hst_0857.pmap' for checking mapping differences.
    CRDS - INFO -  Checking diffs from 'hst_synphot_thruput_0025.rmap' to 'synphot_thruput.rmap'
    CRDS - INFO -  ########################################
    CRDS - INFO -  0 errors
    CRDS - INFO -  0 warnings
    CRDS - INFO -  5 infos
    0

    >>> test_config.cleanup(old_state)
    """

def dt_synphot_refactor():
    """
    >>> old_state = test_config.setup()

    >>> r = crds.get_cached_mapping("data/synphot_thermal.rmap")
    >>> header = dict(COMPNAME="NIC1_BEND", CREATED="2002-03-06 04:51:00",  DESCRIP="Use NIC2 values")

    >>> v = log.set_verbose(55)

    For classic irrelevant parkey matching where values are normalized to N/A and ignored

    >>> print(log.PP(r.map_irrelevant_parkeys_to_na(header)))
    CRDS - DEBUG -  Parkey synphot thermal date is relevant: False 'keep_comments'
    CRDS - DEBUG -  Setting irrelevant parkey 'DATE' to N/A
    CRDS - DEBUG -  Parkey synphot thermal descrip is relevant: False 'keep_comments'
    CRDS - DEBUG -  Setting irrelevant parkey 'DESCRIP' to N/A
    {'COMPNAME': 'NIC1_BEND',
     'CREATED': '2002-03-06 04:51:00',
     'DATE': 'N/A',
     'DESCRIP': 'N/A'}

    For rmap updates which add comment keywords to the match tuples for web display

    >>> print(log.PP(r.map_irrelevant_parkeys_to_na(header, keep_comments=True)))
    CRDS - DEBUG -  Parkey synphot thermal date is relevant: True 'keep_comments'
    CRDS - DEBUG -  Parkey synphot thermal descrip is relevant: True 'keep_comments'
    {'COMPNAME': 'NIC1_BEND',
     'CREATED': '2002-03-06 04:51:00',
     'DESCRIP': 'Use NIC2 values'}

    >>> _ = log.set_verbose(v)

    TMC   rmap

    >>> refactor.RefactorScript("crds.refactor insert data/synphot_tmctab.rmap /tmp/synphot_tmctab.test.rmap data/43h1909cm_tmc.fits")()
    CRDS - INFO -  Inserting 43h1909cm_tmc.fits into 'hst_synphot_tmctab_0021.rmap'
    CRDS - INFO -  0 errors
    CRDS - INFO -  0 warnings
    CRDS - INFO -  1 infos
    0

    >>> diff.DiffScript("crds.diff  data/synphot_tmctab.rmap /tmp/synphot_tmctab.test.rmap")()
    (('data/synphot_tmctab.rmap', '/tmp/synphot_tmctab.test.rmap'), 'replaced 5182153pm_tmc.fits with 43h1909cm_tmc.fits')
    1

    TMG   rmap

    >>> refactor.RefactorScript("crds.refactor insert data/synphot_tmgtab.rmap /tmp/synphot_tmgtab.test.rmap data/2381905mm_tmg.fits")()
    CRDS - INFO -  Inserting 2381905mm_tmg.fits into 'hst_synphot_tmgtab_0007.rmap'
    CRDS - INFO -  0 errors
    CRDS - INFO -  0 warnings
    CRDS - INFO -  1 infos
    0

    >>> diff.DiffScript("crds.diff  data/synphot_tmgtab.rmap /tmp/synphot_tmgtab.test.rmap")()
    (('data/synphot_tmgtab.rmap', '/tmp/synphot_tmgtab.test.rmap'), 'replaced 4cm1612bm_tmg.fits with 2381905mm_tmg.fits')
    1

    TMT   rmap

    >>> refactor.RefactorScript("crds.refactor insert data/synphot_tmttab.rmap /tmp/synphot_tmttab.test.rmap data/tae17277m_tmt.fits")()
    CRDS - INFO -  Inserting tae17277m_tmt.fits into 'hst_synphot_tmttab_0001.rmap'
    CRDS - INFO -  0 errors
    CRDS - INFO -  0 warnings
    CRDS - INFO -  1 infos
    0

    >>> diff.DiffScript("crds.diff  data/synphot_tmttab.rmap /tmp/synphot_tmttab.test.rmap")()
    (('data/synphot_tmttab.rmap', '/tmp/synphot_tmttab.test.rmap'), 'replaced 3241637sm_tmt.fits with tae17277m_tmt.fits')
    1

    THERMAL rmap

    >>> refactor.RefactorScript("crds.refactor insert data/synphot_thermal.rmap /tmp/synphot_thermal.test.rmap data/wfc3_ir_f098m_002_th.fits")()
    CRDS - INFO -  Inserting wfc3_ir_f098m_002_th.fits into 'hst_synphot_thermal_0002.rmap'
    CRDS - INFO -  0 errors
    CRDS - INFO -  0 warnings
    CRDS - INFO -  1 infos
    0

    >>> diff.DiffScript("crds.diff  data/synphot_thermal.rmap /tmp/synphot_thermal.test.rmap")()
    (('data/synphot_thermal.rmap', '/tmp/synphot_thermal.test.rmap'), ('WFC3_IR_F098M', '2006-08-15 08:00:00', 'Filter transmission for F098M'), 'deleted Match rule for wfc3_ir_f098m_002_th.fits')
    (('data/synphot_thermal.rmap', '/tmp/synphot_thermal.test.rmap'), ('WFC3_IR_F098M', '2019-04-02T15:00:35', 'UPDATED TO CONVERT AIR WAVELENGTHS TO VACUUM.'), 'added Match rule for wfc3_ir_f098m_002_th.fits')
    1

    THRUPUT rmap

    >>> refactor.RefactorScript("crds.refactor insert data/synphot_thruput.rmap /tmp/synphot_thruput.test.rmap data/wfc3_uvis_f469nf2_003_syn.fits")()
    CRDS - INFO -  Inserting wfc3_uvis_f469nf2_003_syn.fits into 'hst_synphot_thruput_0025.rmap'
    CRDS - INFO -  0 errors
    CRDS - INFO -  0 warnings
    CRDS - INFO -  1 infos
    0

    >>> diff.DiffScript("crds.diff  data/synphot_thruput.rmap /tmp/synphot_thruput.test.rmap")()
    (('data/synphot_thruput.rmap', '/tmp/synphot_thruput.test.rmap'), ('WFC3_UVIS_F469NF2', '2016-04-20 20:10:00', 'normalization of f469n flat chip 2---------------------------------'), 'deleted Match rule for wfc3_uvis_f469nf2_003_syn.fits')
    (('data/synphot_thruput.rmap', '/tmp/synphot_thruput.test.rmap'), ('WFC3_UVIS_F469NF2', '2019-04-02T15:00:44', 'NORMALIZATION OF F469N FLAT CHIP 2---------------------------------'), 'added Match rule for wfc3_uvis_f469nf2_003_syn.fits')
    1

    >>> test_config.cleanup(old_state)
    """

def dt_synphot_bestrefs():
    """
    >>> old_state = test_config.setup()

    >>> r = crds.get_cached_mapping("data/synphot_thermal.rmap")
    >>> r.get_best_references({'COMPNAME':'NIC1_F110W', 'CREATED':'NO MATCH', 'DESCRIP':'NO MATCH'})
    CRDS - WARNING -  Failed checking OMIT for 'synphot' 'thermal' with expr 'true' : name 'true' is not defined
    CRDS - WARNING -  Failed checking OMIT for 'synphot' 'thermal' with expr 'true' : name 'true' is not defined
    {'thermal': 'nic1_f110w_002_th.fits'}

    >>> r = crds.get_cached_mapping("data/synphot_thruput.rmap")
    >>> r.get_best_references({'COMPNAME':'ACS_BLOCK1', 'CREATED':'NO MATCH', 'DESCRIP':'NO MATCH'})
    CRDS - WARNING -  Failed checking OMIT for 'synphot' 'thruput' with expr 'true' : name 'true' is not defined
    CRDS - WARNING -  Failed checking OMIT for 'synphot' 'thruput' with expr 'true' : name 'true' is not defined
    {'thruput': 'acs_block1_002_syn.fits'}

    >>> test_config.cleanup(old_state)
    """


def dt_synphot_diff():
    """
    >>> old_state = test_config.setup()

    TMC   reference

    >>> diff.DiffScript("crds.diff data/43h1240om_tmc.fits data/43h1909cm_tmc.fits")()   # doctest: +ELLIPSIS
    <BLANKLINE>
     fitsdiff: ...
     a: data/43h1240om_tmc.fits
     b: data/43h1909cm_tmc.fits
     Maximum number of different data values to be reported: 10
     Relative tolerance: 0.0, Absolute tolerance: 0.0
    <BLANKLINE>
    Primary HDU:
    <BLANKLINE>
       Headers contain differences:
         Headers have different number of cards:
          a: 116
          b: 120
         Inconsistent duplicates of keyword 'HISTORY':
          Occurs 106 time(s) in a, 110 times in (b)
         Keyword USEAFTER has different values:
            a> Mar 17 2020 08:40:38
             ?              ^  ----
            b> Mar 17 2020 15:09:14
             ?             +++ ^ +
    <BLANKLINE>
    Extension HDU 1:
    <BLANKLINE>
       Data contains differences:
         Column COMMENT data differs in row 88:
            a> acs/sbc encircled energy table
            b> encircled energy table for ACS/SBC
         Column FILENAME data differs in row 88:
            a> cracscomp$acs_sbc_aper_003_syn.fits[aper#]
             ?                          ^
            b> cracscomp$acs_sbc_aper_004_syn.fits[aper#]
             ?                          ^
         Column TIME data differs in row 88:
            a> Apr 30 2018 11:10:11
            b> Mar 17 2020 15:09:14
         3 different table data element(s) found (0.03% different).
    <BLANKLINE>
         HDU extension #1 contains no differences
    ...

    >>> test_config.cleanup(old_state)
    """

def dt_synphot_core_integration_test():
    """
    >>> old_state = test_config.setup()
    >>> from crds.misc.synphot import SynphotCoreIntegrationTest

    Passing tests

    >>> test = SynphotCoreIntegrationTest("hst_0779.pmap")
    >>> test.run()
    CRDS - INFO -  Checking for components present in TMG COMPNAME but missing from TMC COMPNAME
    CRDS - INFO -  Checking for components present in TMC COMPNAME but missing from TMG COMPNAME
    CRDS - INFO -  Checking for components present in TMG THCOMPNAME but missing from TMT COMPNAME
    CRDS - INFO -  Checking for components present in TMT COMPNAME but missing from TMG THCOMPNAME
    CRDS - INFO -  Checking for components present in TMG COMPNAME but missing from throughput table COMPNAME
    CRDS - INFO -  Checking for components present in throughput table COMPNAME but missing from TMG COMPNAME
    CRDS - INFO -  Checking for components present in TMG THCOMPNAME but missing from thermal table COMPNAME
    CRDS - INFO -  Checking for components present in thermal table COMPNAME but missing from TMG THCOMPNAME
    CRDS - INFO -  Confirming correctly formed TMC filenames
    CRDS - INFO -  Confirming correctly formed TMT filenames
    True

    Components present in graph but missing from lookup/component files

    >>> test = SynphotCoreIntegrationTest("hst_0779.pmap", graph_file="data/hst_synphot_extra_tmg.fits")
    >>> test.run()
    CRDS - INFO -  Checking for components present in TMG COMPNAME but missing from TMC COMPNAME
    CRDS - ERROR -  Components present in TMG COMPNAME but missing from TMC COMPNAME: xtra_thruput_comp
    CRDS - INFO -  Checking for components present in TMC COMPNAME but missing from TMG COMPNAME
    CRDS - INFO -  Checking for components present in TMG THCOMPNAME but missing from TMT COMPNAME
    CRDS - ERROR -  Components present in TMG THCOMPNAME but missing from TMT COMPNAME: xtra_thermal_comp
    CRDS - INFO -  Checking for components present in TMT COMPNAME but missing from TMG THCOMPNAME
    CRDS - INFO -  Checking for components present in TMG COMPNAME but missing from throughput table COMPNAME
    CRDS - ERROR -  Components present in TMG COMPNAME but missing from throughput table COMPNAME: xtra_thruput_comp
    CRDS - INFO -  Checking for components present in throughput table COMPNAME but missing from TMG COMPNAME
    CRDS - INFO -  Checking for components present in TMG THCOMPNAME but missing from thermal table COMPNAME
    CRDS - ERROR -  Components present in TMG THCOMPNAME but missing from thermal table COMPNAME: xtra_thermal_comp
    CRDS - INFO -  Checking for components present in thermal table COMPNAME but missing from TMG THCOMPNAME
    CRDS - INFO -  Confirming correctly formed TMC filenames
    CRDS - INFO -  Confirming correctly formed TMT filenames
    False

    Components missing from graph but present in lookup/component files

    >>> test = SynphotCoreIntegrationTest("hst_0779.pmap", graph_file="data/hst_synphot_missing_tmg.fits")
    >>> test.run()
    CRDS - INFO -  Checking for components present in TMG COMPNAME but missing from TMC COMPNAME
    CRDS - INFO -  Checking for components present in TMC COMPNAME but missing from TMG COMPNAME
    CRDS - ERROR -  Components present in TMC COMPNAME but missing from TMG COMPNAME: wfirst_wfi_prism
    CRDS - INFO -  Checking for components present in TMG THCOMPNAME but missing from TMT COMPNAME
    CRDS - INFO -  Checking for components present in TMT COMPNAME but missing from TMG THCOMPNAME
    CRDS - INFO -  Checking for components present in TMG COMPNAME but missing from throughput table COMPNAME
    CRDS - INFO -  Checking for components present in throughput table COMPNAME but missing from TMG COMPNAME
    CRDS - ERROR -  Components present in throughput table COMPNAME but missing from TMG COMPNAME: wfirst_wfi_prism
    CRDS - INFO -  Checking for components present in TMG THCOMPNAME but missing from thermal table COMPNAME
    CRDS - INFO -  Checking for components present in thermal table COMPNAME but missing from TMG THCOMPNAME
    CRDS - INFO -  Confirming correctly formed TMC filenames
    CRDS - INFO -  Confirming correctly formed TMT filenames
    False

    Invalid filename in throughput lookup table

    >>> test = SynphotCoreIntegrationTest("hst_0779.pmap", throughput_lookup_file="data/hst_synphot_malformed_filename_tmc.fits")
    >>> test.run()
    CRDS - INFO -  Checking for components present in TMG COMPNAME but missing from TMC COMPNAME
    CRDS - INFO -  Checking for components present in TMC COMPNAME but missing from TMG COMPNAME
    CRDS - INFO -  Checking for components present in TMG THCOMPNAME but missing from TMT COMPNAME
    CRDS - INFO -  Checking for components present in TMT COMPNAME but missing from TMG THCOMPNAME
    CRDS - INFO -  Checking for components present in TMG COMPNAME but missing from throughput table COMPNAME
    CRDS - INFO -  Checking for components present in throughput table COMPNAME but missing from TMG COMPNAME
    CRDS - INFO -  Checking for components present in TMG THCOMPNAME but missing from thermal table COMPNAME
    CRDS - INFO -  Checking for components present in thermal table COMPNAME but missing from TMG THCOMPNAME
    CRDS - INFO -  Confirming correctly formed TMC filenames
    CRDS - ERROR -  Malformed TMC filename, expected 'crwfpccomp$wfpc_pol60_003_syn.fits', found 'crwfpccomp$wfpc_foo_003_syn.fits'
    CRDS - INFO -  Confirming correctly formed TMT filenames
    False

    Invalid filename in thermal lookup table

    >>> test = SynphotCoreIntegrationTest("hst_0779.pmap", thermal_lookup_file="data/hst_synphot_malformed_filename_tmt.fits")
    >>> test.run()
    CRDS - INFO -  Checking for components present in TMG COMPNAME but missing from TMC COMPNAME
    CRDS - INFO -  Checking for components present in TMC COMPNAME but missing from TMG COMPNAME
    CRDS - INFO -  Checking for components present in TMG THCOMPNAME but missing from TMT COMPNAME
    CRDS - INFO -  Checking for components present in TMT COMPNAME but missing from TMG THCOMPNAME
    CRDS - INFO -  Checking for components present in TMG COMPNAME but missing from throughput table COMPNAME
    CRDS - INFO -  Checking for components present in throughput table COMPNAME but missing from TMG COMPNAME
    CRDS - INFO -  Checking for components present in TMG THCOMPNAME but missing from thermal table COMPNAME
    CRDS - INFO -  Checking for components present in thermal table COMPNAME but missing from TMG THCOMPNAME
    CRDS - INFO -  Confirming correctly formed TMC filenames
    CRDS - INFO -  Confirming correctly formed TMT filenames
    CRDS - ERROR -  Malformed TMT filename, expected 'crwfc3comp$wfc3_ir_f167n_002_th.fits', found 'crwfc3comp$wfc3_ir_foo_002_th.fits'
    False
    """

def dt_synphot_obsmodes_integration_test():
    """
    >>> old_state = test_config.setup()
    >>> from crds.misc.synphot import SynphotObsmodeIntegrationTest

    Passing tests

    >>> test = SynphotObsmodeIntegrationTest(
    ... "hst_0779.pmap",
    ... "data/synphot_root",
    ... "data/hst_synphot_passing_obs.fits"
    ... )
    >>> test.run()
    CRDS - INFO -  Creating bandpass objects from 1 observation modes
    CRDS - INFO -  Congratulations, all observation modes succeeded!
    True

    Passing tests in multiprocessing mode

    >>> test = SynphotObsmodeIntegrationTest(
    ... "hst_0779.pmap",
    ... "data/synphot_root",
    ... "data/hst_synphot_passing_obs.fits",
    ... processes=4
    ... )
    >>> test.run()
    CRDS - INFO -  Creating bandpass objects from 1 observation modes
    CRDS - INFO -  Congratulations, all observation modes succeeded!
    True

    Failure due to missing component files

    >>> test = SynphotObsmodeIntegrationTest(
    ... "hst_0779.pmap",
    ... "data/synphot_root",
    ... obsmodes_file="data/hst_synphot_failing_obs.fits"
    ... )
    >>> test.run()
    CRDS - INFO -  Creating bandpass objects from 1 observation modes
    CRDS - ERROR -  Exception from stsynphot with obsmode 'acs,hrc,coron': FileNotFoundError(2, 'No such file or directory')
    CRDS - ERROR -  Exception from pysynphot with obsmode 'acs,hrc,coron': FileNotFoundError(2, 'No such file or directory')
    CRDS - INFO -  1 / 1 observation modes failed
    False

    Failure in multiprocessing mode

    >>> test = SynphotObsmodeIntegrationTest(
    ... "hst_0779.pmap",
    ... "data/synphot_root",
    ... obsmodes_file="data/hst_synphot_failing_obs.fits",
    ... processes=4
    ... )
    >>> test.run()
    CRDS - INFO -  Creating bandpass objects from 1 observation modes
    CRDS - ERROR -  Exception from stsynphot with obsmode 'acs,hrc,coron': FileNotFoundError(2, 'No such file or directory')
    CRDS - ERROR -  Exception from pysynphot with obsmode 'acs,hrc,coron': FileNotFoundError(2, 'No such file or directory')
    CRDS - INFO -  1 / 1 observation modes failed
    False

    """

# ==============================================================

def main():
    """Run module tests,  for now just doctests only.

    test_config.setup() and cleanup() are done inline above because bracketing
    the tests here does not get picked up by nose test discovery.  Combining
    tests into one giant docstring works but is hard to analyze and debug when
    things go wrong.
    """
    from crds.tests import test_synphot_hst, tstmod
    return tstmod(test_synphot_hst)

# ==============================================================

if __name__ == "__main__":
    print(main())



"""This module contains doctests and unit tests which exercise refactoring
code,  modules used to automatically update mappings.

"""
import os
from pprint import pprint as pp

import crds
from crds.core import log, exceptions
from crds import data_file, diff

from crds.diff import DiffScript
from crds.bestrefs import BestrefsScript
from crds.refactoring import refactor
from crds.refactoring.refactor import RefactorScript
from crds.certify import CertifyScript
from crds.tests import test_config

from nose.tools import assert_raises, assert_true

# ==================================================================================

def dt_or_bars_certify_bad_keyword():
    """
    >>> old_state = test_config.setup(url="https://jwst-serverless-mode.stsci.edu")
    >>> CertifyScript("crds.certify data/jwst_miri_ipc.bad-keyword.fits --comparison-context jwst_0361.pmap")() # doctest: +ELLIPSIS
    CRDS - INFO -  ########################################
    CRDS - INFO -  Certifying 'data/jwst_miri_ipc.bad-keyword.fits' (1/1) as 'FITS' relative to context 'jwst_0361.pmap'
    CRDS - INFO -  FITS file 'jwst_miri_ipc.bad-keyword.fits' conforms to FITS standards.
    CRDS - WARNING -  CRDS-pattern-like keyword 'P_DETEC' w/o CRDS translation to corresponding dataset keyword.
    CRDS - INFO -  Pattern-like keyword 'P_DETEC' may be misspelled or missing its translation in CRDS.  Pattern will not be used.
    CRDS - INFO -  The translation for 'P_DETEC' can be defined in crds.jwst.locate or rmap header reference_to_dataset field.
    CRDS - INFO -  If this is not a pattern keyword, adding a translation to 'not-a-pattern' will suppress this warning.
    CRDS - INFO -  Checking JWST datamodels.
    CRDS - INFO -  ########################################
    CRDS - INFO -  0 errors
    CRDS - INFO -  1 warnings
    CRDS - INFO -  8 infos
    0
    >>> test_config.cleanup(old_state)
    """

def dt_or_bars_certify_bad_value():
    """
    >>> old_state = test_config.setup(url="https://jwst-serverless-mode.stsci.edu")
    >>> CertifyScript("crds.certify data/jwst_miri_ipc.bad-value.fits --comparison-context jwst_0361.pmap")() # doctest: +ELLIPSIS
    CRDS - INFO -  ########################################
    CRDS - INFO -  Certifying 'data/jwst_miri_ipc.bad-value.fits' (1/1) as 'FITS' relative to context 'jwst_0361.pmap'
    CRDS - INFO -  FITS file 'jwst_miri_ipc.bad-value.fits' conforms to FITS standards.
    CRDS - INFO -  Setting 'META.INSTRUMENT.BAND [BAND]' = None to value of 'P_BAND' = 'LONG'
    CRDS - INFO -  Setting 'META.INSTRUMENT.DETECTOR [DETECTOR]' = 'MIRIMAGE' to value of 'P_DETECT' = 'MIRIFUSHORT|FOO|'
    CRDS - ERROR -  instrument='MIRI' type='IPC' data='data/jwst_miri_ipc.bad-value.fits' ::  Checking 'META.INSTRUMENT.DETECTOR [DETECTOR]' : Value 'FOO' is not one of ['ANY', 'MIRIFULONG', 'MIRIFUSHORT', 'MIRIMAGE', 'N/A']
    CRDS - INFO -  Checking JWST datamodels.
    CRDS - INFO -  ########################################
    CRDS - INFO -  1 errors
    CRDS - INFO -  0 warnings
    CRDS - INFO -  7 infos
    1
    >>> test_config.cleanup(old_state)
    """

def dt_or_bars_refactor_add_file():
    """
    >>> old_state = test_config.setup(url="https://jwst-serverless-mode.stsci.edu")
    >>> RefactorScript("crds.refactor insert data/jwst_miri_ipc_0002.rmap ./jwst_miri_ipc_0003.add.rmap data/jwst_miri_ipc_0003.add.fits")()  # doctest: +ELLIPSIS
    CRDS - INFO -  Inserting jwst_miri_ipc_0003.add.fits into 'jwst_miri_ipc_0002.rmap'
    CRDS - INFO -  Setting 'META.INSTRUMENT.BAND [BAND]' = 'UNDEFINED' to value of 'P_BAND' = 'SHORT | MEDIUM |'
    CRDS - INFO -  Setting 'META.INSTRUMENT.DETECTOR [DETECTOR]' = 'MIRIMAGE' to value of 'P_DETECT' = 'MIRIFUSHORT|MIRIFULONG|'
    CRDS - INFO -  0 errors
    CRDS - INFO -  0 warnings
    CRDS - INFO -  3 infos
    0

    >>> diff.DiffScript("crds.diff data/jwst_miri_ipc_0002.rmap ./jwst_miri_ipc_0003.add.rmap")()
    (('data/jwst_miri_ipc_0002.rmap', './jwst_miri_ipc_0003.add.rmap'), ('MIRIFULONG|MIRIFUSHORT', 'MEDIUM|SHORT'), ('2014-01-01', '00:00:00'), 'added Match rule for jwst_miri_ipc_0003.add.fits')
    1

    >>> pp(refactor.rmap_check_modifications("data/jwst_miri_ipc_0002.rmap", "./jwst_miri_ipc_0003.add.rmap", "none", "data/jwst_miri_ipc_0003.add.fits", expected=("add_rule",)))
    True

    >>> _ = os.system("rm ./*.rmap")

    >>> test_config.cleanup(old_state)
    """

def dt_or_bars_refactor_replace_file():
    """
    >>> old_state = test_config.setup(url="https://jwst-serverless-mode.stsci.edu")
    >>> RefactorScript("crds.refactor insert data/jwst_miri_ipc_0002.rmap ./jwst_miri_ipc_0004.replace.rmap data/jwst_miri_ipc_0004.replace.fits")()  # doctest: +ELLIPSIS
    CRDS - INFO -  Inserting jwst_miri_ipc_0004.replace.fits into 'jwst_miri_ipc_0002.rmap'
    CRDS - INFO -  Setting 'META.INSTRUMENT.BAND [BAND]' = 'UNDEFINED' to value of 'P_BAND' = 'LONG |'
    CRDS - INFO -  Setting 'META.INSTRUMENT.DETECTOR [DETECTOR]' = 'MIRIMAGE' to value of 'P_DETECT' = 'MIRIFULONG|'
    CRDS - INFO -  0 errors
    CRDS - INFO -  0 warnings
    CRDS - INFO -  3 infos
    0

    >>> diff.DiffScript("crds.diff data/jwst_miri_ipc_0002.rmap ./jwst_miri_ipc_0004.replace.rmap")()
    (('data/jwst_miri_ipc_0002.rmap', './jwst_miri_ipc_0004.replace.rmap'), ('MIRIFULONG', 'LONG'), ('1900-01-01', '00:00:00'), 'replaced jwst_miri_ipc_0004.fits with jwst_miri_ipc_0004.replace.fits')
    1

    >>> pp(refactor.rmap_check_modifications("data/jwst_miri_ipc_0002.rmap", "./jwst_miri_ipc_0004.replace.rmap", "jwst_miri_ipc_0004.fits", "data/jwst_miri_ipc_0004.replace.fits", expected=("replace",)))
    True

    >>> refactor.rmap_check_modifications("data/jwst_miri_ipc_0002.rmap", "./jwst_miri_ipc_0004.replace.rmap", "data/jwst_miri_ipc_0004.fits", "data/jwst_miri_ipc_0004.replace.fits",
    ...                          expected=("replace",))
    True

    >>> _ = os.system("rm ./*.rmap")

    >>> test_config.cleanup(old_state)
    """

# ==================================================================================


def tst():
    """Run module tests,  for now just doctests only."""
    from crds.tests import test_or_bars, tstmod
    return tstmod(test_or_bars)

if __name__ == "__main__":
    print(tst())



"""This module manages the formal specification of reference file constraints
as a generic system.

The name TPN is a legacy term from the predecessor of CRDS, HST's CDBS.  It's
the file extension for files used to describe the constraints for a specific
instrument and type.

TPN files are unique to each project/telescope and are defined in the
observatory package in the "tpns" subdirectory.    Each TPN file defines
a number of constraints for references.   Each individual constraint is
represented in CRDS by a named tuple called a TpnInfo.  Each TpnInfo later
becomes the core specification of a Validator object that verifies one
file constraint.  The CRDS certifier applies a list of Validator objects
to reference files as the central mechanism for constraint checking.

Most HST .tpn files were inherited unchanged from CDBS.  HST .tpn files come
in two forms:

<instrument> _ <type_suffix> .tpn     (constrain reference files)
<instrument> _ <type_suffix> _ld.tpn  (constrain database, now rmaps)

In rare cases,  HST specializes .tpn files based on header keywords other
than instrument and type found in the reference file.  In general,  for
HST each instrument+type pair is defined.

JWST extensions reduce duplication in the specifications by extending the
generic .tpn framework to support these additional forms:

all _ all .tpn            (contraints for all instruments and types)
<instrument> _ all .tpn   (constraints for all types of one instrument)
all _ <type_suffix> .tpn  (constraints for all instruments for one type)

For JWST generally .tpn and _ld.tpn's are identical.  For JWST, most checks are
done using the extended forms.  As a matter of specification, all the .tpn
files are enforced equally; however, typically the more specialized forms of
specification (specific instrument, type, or both) define tighter constraints,
so there is an assumed process of narrowing.  e.g. Keywords defined as
"optional" at more generic levels may become required, and values only defined
broadly may be narrowed to those applicable to a specific instrument or type,
etc.

In addition, JWST processes the calibration code data model schema and converts
schema entries into optional TpnInfo objects; this enables CRDS to expose the
checks being applied by the data model and apply them directly in CRDS as well.
"""
import os.path
import collections
import re

# ============================================================================

from crds.core import log, utils, exceptions
#     from crds.core import rmap, heavy_client    # deferred

# ============================================================================

#
# Only the first character of the field is stored, i.e. Header == H
#
# name = field identifier
# keytype = (Header|Group|Column)
# datatype = (Integer|Real|Logical|Double|Character)
# presence = (Optional|Required)
# values = [...]
#
_TpnInfo = collections.namedtuple("TpnInfo", "name,keytype,datatype,presence,values")

class TpnInfo(_TpnInfo):
    """Named tuple describing a file checking constraint with enhanced repr()."""

    def __repr__(self):
        return ("(" + repr(self.name) + ", "
                + self._repr_keytype() + ", "
                + self._repr_datatype() + ", "
                + self._repr_presence() + ", "
                + self._repr_values() + ")")

    keytypes = {
        "H" : "HEADER",
        "C" : "COLUMN",
        "G" : "GROUP",
        "A" : "ARRAY_FORMAT",
        "D" : "ARRAY_DATA",
        "X" : "EXPRESSION",
    }

    def _repr_keytype(self):
        return repr(self.keytypes.get(self.keytype[0], self.keytype[0]))

    datatypes = {
        "C" : "CHARACTER",
        "S" : "CASE_SENSITIVE_CHARACTER",
        "I" : "INTEGER",
        "L" : "LOGICAL",
        "R" : "REAL",
        "D" : "DOUBLE",
        "X" : "EXPRESSION",
    }

    def _repr_datatype(self):
        return repr(self.datatypes.get(self.datatype[0], self.datatype[0]))

    presences = {
        "E" : "EXCLUDED",
        "R" : "REQUIRED",
        "P" : "REQUIRED",
        "W" : "WARN",
        "O" : "OPTIONAL",
        "F" : "IF_FULL_FRAME",
        "S" : "IF_SUBARRAY",
        "A" : "ANY_SUBARRAY"
    }

    def _repr_presence(self):
        if is_expression(self.presence):
            return "condition="+repr(self.presence)
        return repr(self.presences.get(self.presence[0], self.presence[0]))

    def _repr_values(self):
        if self.values and is_expression(self.values[0]):
            return "expression=" + repr(self.values[0])
        else:
            return "values=" + repr(self.values)

    @property
    def is_expression(self):
        """Return True IFF this is an expression constraint."""
        return self.datatype[0] == "X"

    @property
    def is_conditionally_applicable(self):
        """Return True IFF this constraint has an expression defining when it is applicable."""
        return is_expression(self.presence)

    @property
    def is_complex_constraint(self):
        """Used to eliminate infos not appropriate as rmap value lists."""
        return self.is_expression or self.is_conditionally_applicable

# =============================================================================

HERE = os.path.dirname(__file__) or "./"

# =============================================================================

def load_tpn(fname):
    """Load a TPN file and return it as a list of TpnInfo objects
    describing keyword requirements including acceptable values.
    """
    tpn = []
    for line in load_tpn_lines(fname):
        line = _fix_quoted_whitespace(line)
        items = line.split()
        items = _restore_embedded_spaces(items)
        items = _remove_quotes(items)
        if len(items) == 4:
            name, keytype, datatype, presence = items
            values = []
        else:
            name, keytype, datatype, presence, values = items
            values = _remove_quotes(values.split(",") if datatype != "X" else [values])
            values = [str(v) if is_expression(v) or datatype == "S" else str(v.upper()) for v in values]
        tpn.append(TpnInfo(name, keytype, datatype, presence, tuple(values)))
    return tpn

def is_expression(tpn_field):
    """Return True IFF .tpn value `tpn_field` defines a header expression.

    Initially `tpn_field` is either a value from the TpnInfo "values" field or
    it is the value of the "presence" field.  In both cases,  an expression
    is signified by bracketing the value in parens.
    """
    return tpn_field.startswith("(") and tpn_field.endswith(")")

@utils.cached
def load_tpn_lines(fname, replacements=()):
    """Load the lines of a CDBS .tpn file,  ignoring #-comments, blank lines,
     and joining lines ending in \\.  If a line begins with "include",  the
    second word should be a base filename that refers to a file in the same
    directory as `fname`.  The lines of the include file are recursively included.
    """
    log.verbose("Loading .tpn lines from", log.srepr(fname),
                "with replacements", log.srepr(replacements), verbosity=80)
    lines = []
    append = False
    dirname = os.path.dirname(fname)
    with open(fname) as pfile:
        for line in pfile:
            line = line.strip()
            if line.startswith("#") or not line:
                continue
            if line.startswith("include"):   #  include tpn_file
                fname2 = os.path.join(dirname, line.split()[1])
                lines += load_tpn_lines(fname2, replacements)
                continue
            elif line.startswith("replace"): #  replace orig_str  new_str
                orig, replaced = replacement = tuple(line.split()[1:])
                if replacement not in replacements:
                    for replacement2 in replacements:
                        orig2, replaced2 = replacement2
                        if orig == orig2 and replaced != replaced2:
                            raise exceptions.InconsistentTpnReplaceError(
                                "In", repr(fname),
                                "Tpn replacement directive", repr(replacement),
                                "conflicts with directive", repr(replacement2))
                    else:
                        replacements = replacements + (replacement,)
                else:
                    log.verbose("Duplicate replacement", replacement, verbosity=80)
                continue
            for (orig, new) in replacements:
                line = re.sub(orig, new, line)
            if append:
                lines[-1] = lines[-1][:-1].strip() + line
            else:
                lines.append(line)
            append = line.endswith("\\")
    return lines

SPACE_MAGIC = "@@1324$$"

def _fix_quoted_whitespace(line):
    """Replace spaces and tabs which appear inside quotes in `line` with
    magic,  and return it.
    """
    i = 0
    while i < len(line):
        char = line[i]
        i += 1
        if char != '"':
            continue
        quote = char
        while i < len(line):
            char = line[i]
            i += 1
            if char == quote:
                break
            if char in " \t":
                line = line[:i-1] + SPACE_MAGIC + line[i:]
    return line

def _restore_embedded_spaces(values):
    """Undo space encoding needed to make simple splits work for TpnInfos."""
    return [value.replace(SPACE_MAGIC, " ") for value in values]

def _remove_quotes(values):
    """Remove any quotes from quoted values."""
    removed = []
    for value in values:
        if value.startswith('"') and value.endswith('"'):
            value = value[1:-1]
        removed.append(value)
    return removed

@utils.cached
def get_tpninfos(filepath):
    """Load the list of TPN info tuples from .tpn file at `filepath`.  Unlike
    load_tpn(), this function is structured such that missing files are an
    expected error and relegated to a verbose warning.  This is because for
    JWST the full pantheon of possible .tpn files is never created.
    """
    # XXXX Doesn't use verbose_warning_on_exception because --debug-traps would
    # always trigger since trapped exceptions are expected here,  debugging this
    # is trickier than normal.
    return load_tpn(filepath) if os.path.exists(filepath) else []

def get_tpn_path(tpn, observatory):
    """Return the absolute path to the `tpn` file belonging to `observatory`."""
    locator = utils.get_locator_module(observatory)
    return locator.tpn_path(tpn)

# =============================================================================

def load_all_type_constraints(observatory):
    """Load all the type constraint files from `observatory` package.

    There are constraints that apply to:

    ALL instruments and types
    ALL types of one instrument
    ALL instruments of one type
    One instrument and type

    Generally these should be thought of as designed for successive refinement,
    so all constraints are applied, but as their scope narrows they can become
    stricter.  Since increasing strictness and refinement require more knowledge,
    the development order of the constraints mirrored that.

    However, in the (revised) loading below, constraints are loaded by order of
    decreasing strictness; this makes it possible to define strict
    constants/replacements early in the loading process and to apply those
    to customize the more generalized constraints loaded later.
    """
    from crds.core import rmap, heavy_client
    pmap_name = heavy_client.load_server_info(observatory).operational_context
    pmap = rmap.get_cached_mapping(pmap_name)
    locator = utils.get_locator_module(observatory)
    for instr in pmap.selections:
        imap = pmap.get_imap(instr)
        for filekind in imap.selections:
            if imap.selections[filekind] == "N/A":
                continue
            try:
                suffix  = locator.TYPES.filekind_to_suffix(instr, filekind)
            except Exception as exc:
                log.warning("Missing suffix coverage for", repr((instr, filekind)), ":", exc)
            else:
                locator.get_all_tpninfos(instr, suffix, "tpn")  # With core schema,  one type loads all
                locator.get_all_tpninfos(instr, suffix, "ld_tpn")  # With core schema,  one type loads all
                locator.get_all_tpninfos("all", suffix, "tpn")  # With core schema,  one type loads all
                locator.get_all_tpninfos("all", suffix, "ld_tpn")  # With core schema,  one type loads all
        locator.get_all_tpninfos(instr, "all", "tpn")
        locator.get_all_tpninfos(instr, "all", "ld_tpn")
    locator.get_all_tpninfos("all","all","tpn")
    locator.get_all_tpninfos("all","all","ld_tpn")

# =============================================================================

def main():
    """Place holder function for running this module as cmd line program."""
    print("null tpn processing.")

if __name__ == "__main__":
    main()



"""This module defines dictionary-like clasess of various forms nominally used to support
CRDS mappings:

TransformedDict --   generic MutableMapping with key and value transforms
    Derived from discussion at http://stackoverflow.com/questions/3387691/python-how-to-perfectly-override-a-dict

LazyFileDict   --  Demand loaded network of files (aka CRDS context)

"""
from collections.abc import MutableMapping

# =============================================================================

class TransformedDict(MutableMapping):

    """A dictionary that applies an arbitrary key and value altering functions
    before accessing the keys and values.
    """

    def __init__(self, initializer=()):
        self._contents = dict()
        self.update(dict(initializer))

    def __getstate__(self):
        return dict(_contents = self._contents)

    def __setstate__(self, state):
        self._contents = state["_contents"]

    def transform_key(self, key):
        """Abstract pass-thru key transformation,  override needed."""
        return key

    def transform_value(self, value):
        """Abstract pass-thru value transformation,  override as needed."""
        return value

    def __getitem__(self, key):
        return self._contents[self.transform_key(key)]

    def __setitem__(self, key, value):
        self._contents[self.transform_key(key)] = self.transform_value(value)

    def __delitem__(self, key):
        del self._contents[self.transform_key(key)]

    def __iter__(self):
        return iter(self._contents)

    def __len__(self):
        return len(self._contents)

    def get(self, key, default=None):
        """Returns either value associated with `key` or transformed `default` value."""
        try:
            return self[key]
        except KeyError:
            return default

    def __repr__(self):
        """
        >>> TransformedDict([("this","THAT"), ("ANOTHER", "(ESCAPED)")])
        TransformedDict([('ANOTHER', '(ESCAPED)'), ('this', 'THAT')])
        """
        return self.__class__.__name__ + "({})".format(sorted(list(self.items())))

# =============================================================================

class LazyFileDict(TransformedDict):
    """Instantiates a graph of inter-referencing files (like CRDS Mappings) on
     demand basis as particular paths are accessed.

    Manages selections for higher level mappings like .pmaps and .imaps which
    refer to other mapping files.  Loads submappings on a demand basis rather
    than exhaustively at startup.  This is motivated by a need to minimize file
    operations in web/cloud applications and highly concurrent environments
    like calibration pipelines.  It opens the potential for only loading a
    single instrument under a given context.

    Values included in special_values_set are excluded from nested loading
    and kept as literals,  e.g. N/A
    """

    special_values_set = set()   # Override in sub-classes as needed

    def __init__(self, selector, load_keys):
        assert "loader" in load_keys
        self._xx_load_keys = load_keys
        self._xx_selector = {}
        super(LazyFileDict, self).__init__()
        self._xx_selector = {self.transform_key(key): value for (key, value) in dict(selector).items()}

    def clear(self):
        """Drop all loaded selections reverting to pre-demand-loaded state."""
        super(LazyFileDict, self).__init__()

    def __getstate__(self):
        """Drop dictionary attributes which correspond to loaded/cached members."""
        return dict(
            _xx_load_keys = self._xx_load_keys,
            _xx_selector = self._xx_selector,
            _xx_superclass = super(LazyFileDict, self).__getstate__()
            )

    def __setstate__(self, state):
        """Drop dictionary attributes which correspond to loaded/cached members."""
        self.__init__(state["_xx_selector"], state["_xx_load_keys"])
        super(LazyFileDict, self).__setstate__(state["_xx_superclass"])

    @classmethod
    def is_special_value(cls, value):
        """Returns True IFF `value` is a string in cls.special_values_set.
        Nominally this means the value has no recursive load operation, it's
        a literal.
        """
        return isinstance(value, str) and value in cls.special_values_set

    def __getitem__(self, name):
        name = self.transform_key(name)
        if name in self._xx_selector:
            try:
                val = super(LazyFileDict, self).__getitem__(name)
            except KeyError:
                if self.is_special_value(self._xx_selector[name]):
                    val = self._xx_selector[name]
                else:
                    val = self._xx_load_keys["loader"](self._xx_selector[name], **self._xx_load_keys)
                super(LazyFileDict, self).__setitem__(name, val)
            return val
        else:
            raise KeyError(name)

    def __setitem__(self, name, value):
        name = self.transform_key(name)
        self._xx_selector[name] = value.filename
        super(LazyFileDict, self).__setitem__(name, value)

    def __delitem__(self, name):
        name = self.transform_key(name)
        del self._xx_selector[name]
        super(LazyFileDict, self).__delitem__(name)

    def __iter__(self):
        return iter(self.keys())

    def __len__(self):
        return len(self._xx_selector)

    def keys(self):
        """
        NOTE:  does not require full load
        """
        return self._xx_selector.keys()

    def normal_keys(self):
        """Each of these keys has a corresponding value which IS NOT special.

        NOTE:  Does not require full load.
        """
        return sorted([key for key in self.keys() if not self.is_special_value(self._xx_selector[key])])

    def special_keys(self):
        """Each of these keys has a corresponding values which IS special.

        NOTE:  Does not require full load.
        """
        return sorted([key for key in self.keys() if self.is_special_value(self._xx_selector[key]) ])

    def values(self):
        """Return all the values of this LazyFileDict,  implicitly loading them all.

        NOTE:  REQUIRES FULL LOAD
        """
        return [self[name] for name in self.keys()]

    def normal_values(self):
        """Normal values exclude the special values like N/A but can include exotic values like tuples or dicsts.

        NOTE:  REQUIRES FULL LOAD
        """
        return [self[key] for key in self.normal_keys()]

    def special_values(self):
        """These are values which must be trapped and reformatted in the Mapping classes."""
        return [self[key] for key in self.special_keys()]

    def items(self):
        """Return all the items of this LazyFileDict, implicitly loading all values.

        NOTE:  REQUIRES FULL LOAD
        """
        return self._items(self.keys())

    def normal_items(self):
        """Return only those items that have recursively loaded values.

        NOTE:  REQUIRES FULL LOAD
        """
        return self._items(self.normal_keys())

    def special_items(self):
        """Return only those items that have exempt literal values with no recursive loading."""
        return self._items(self.special_keys())

    def _items(self, keys):
        """Constructs item list taken from self based on `keys`."""
        return [(key, self[key]) for key in sorted(keys)]

    def __repr__(self):
        """Override __repr__ to prevent forced load of all selector items just for __repr__."""
        return self.__class__.__name__ + "(" + repr(sorted(self._xx_selector.items())) + \
            ", " + repr(sorted(self._xx_load_keys.items())) + ")"



import os
import datetime

from astropy.io import fits
import numpy as np

from crds.core import rmap
from crds import data_file

from . import utils


_TIMESTAMP_FORMAT = "%b %d %Y %H:%M:%S"

_AM_PM_TIMESTAMP_FORMAT = "%b %d %Y %I:%M:%S:%f%p"

_PEDIGREE_DATE_FORMAT = "%d/%m/%Y"

_DBTABLE_BY_REFTYPE = {
    utils.THROUGHPUT_LOOKUP_REFTYPE: "CRCOMPLIST",
    utils.THERMAL_LOOKUP_REFTYPE: "CRTHERMLIST",
}

_DESCRIP_BY_REFTYPE = {
    utils.THROUGHPUT_LOOKUP_REFTYPE: "The lookup table for HST component throughputs",
    utils.THERMAL_LOOKUP_REFTYPE: "The lookup table for HST component thermal characteristics",
}

_HISTORY_DESCRIPTION_BY_REFTYPE = {
    utils.THROUGHPUT_LOOKUP_REFTYPE: "throughput",
    utils.THERMAL_LOOKUP_REFTYPE: "thermal",
}


class SynphotLookupGenerator:
    """
    Given a derived-from context and a set of newly delivered synphot files,
    produce an updated lookup file.
    """
    def __init__(self, context):
        self._context = context

    def generate(self, reftype, delivered_files):
        """
        Generate and return an HDUList containing an updated lookup table
        for the specified reftype.
        """
        # Historically ReDCaT has always set this to the current
        # local time and not UTC.
        timestamp = datetime.datetime.today().strftime(_TIMESTAMP_FORMAT)

        original_ref_path = utils.get_cache_path(self._context, reftype, error_on_missing=False)
        if original_ref_path is None:
            return self._generate_hdul(reftype, delivered_files, timestamp)
        else:
            with data_file.fits_open(original_ref_path) as original_ref:
                return self._generate_hdul(reftype, delivered_files, timestamp, original_ref=original_ref)

    def _generate_hdul(self, reftype, delivered_files, timestamp, original_ref=None):
        if original_ref is not None:
            original_table = original_ref[-1].data
            new_time = original_table["TIME"]
            new_compname = original_table["COMPNAME"]
            new_filename = original_table["FILENAME"]
            new_comment = original_table["COMMENT"]
        else:
            new_time = np.chararray((0,), unicode=True)
            new_compname = np.chararray((0,), unicode=True)
            new_filename = np.chararray((0,), unicode=True)
            new_comment = np.chararray((0,), unicode=True)

        updated_instruments = set()
        created_instruments = set()
        for file in delivered_files:
            with data_file.fits_open(file) as hdul:
                component = hdul[0].header["COMPNAME"]
                lookup_filename = utils.get_lookup_filename(component, os.path.basename(file))
                description = hdul[0].header.get("DESCRIP", "")
                if component in new_compname:
                    idx = np.argwhere(new_compname == component)[0][0]
                    new_time[idx] = timestamp
                    new_filename[idx] = lookup_filename
                    new_comment[idx] = description
                    updated_instruments.add(utils.get_instrument(component))
                else:
                    new_time = np.append(new_time, timestamp)
                    new_compname = np.append(new_compname, component)
                    new_filename = np.append(new_filename, lookup_filename)
                    new_comment = np.append(new_comment, description)
                    created_instruments.add(utils.get_instrument(component))

        ind = np.argsort(new_compname)
        columns = [
            fits.Column(name="TIME", format="A26", array=new_time[ind], disp="A26"),
            fits.Column(name="COMPNAME", format="A18", array=new_compname[ind], disp="A18"),
            fits.Column(name="FILENAME", format="A56", array=new_filename[ind], disp="A56"),
            fits.Column(name="COMMENT", format="A68", array=new_comment[ind], disp="A68")
        ]

        new_table_hdu = fits.BinTableHDU.from_columns(columns)

        new_primary_hdu = fits.PrimaryHDU()
        header = new_primary_hdu.header
        header["USEAFTER"] = timestamp
        header["INSTRUME"] = "HST"
        header["COMMENT"] = "Reference file automatically generated by CRDS"
        header["DBTABLE"] = _DBTABLE_BY_REFTYPE[reftype]
        header["DESCRIP"] = _DESCRIP_BY_REFTYPE[reftype]
        header["PEDIGREE"] = self._make_pedigree(new_table_hdu)

        if original_ref:
            for item in original_ref[0].header["HISTORY"]:
                header.add_history(item)

            header.add_history(" ")
            header.add_history("Updated on {}".format(timestamp))
        else:
            header.add_history("Created on {}".format(timestamp))

        header.add_history(" ")
        self._add_instrument_history(reftype, header, updated_instruments, created_instruments)

        return fits.HDUList([new_primary_hdu, new_table_hdu])

    def _add_instrument_history(self, reftype, header, updated_instruments, created_instruments):
        description = _HISTORY_DESCRIPTION_BY_REFTYPE[reftype]

        if len(updated_instruments) > 0:
            instrument_list = self._make_instrument_list(updated_instruments)
            header.add_history("Updated {} tables for {}".format(description, instrument_list))

        if len(created_instruments) > 0:
            instrument_list = self._make_instrument_list(created_instruments)
            header.add_history("Created {} tables for {}".format(description, instrument_list))

    def _make_instrument_list(self, instruments):
        instruments = sorted([i.upper() for i in instruments])
        if len(instruments) == 1:
            return instruments[0]
        elif len(instruments) == 2:
            return instruments[0] and instruments[1]
        else:
            return ", ".join(instruments[:-1]) + ", and " + instruments[-1]

    def _make_pedigree(self, table_hdu):
        times = [self._parse_time(t) for t in table_hdu.data["TIME"]]
        return "INFLIGHT {} {}".format(
            min(times).strftime(_PEDIGREE_DATE_FORMAT),
            max(times).strftime(_PEDIGREE_DATE_FORMAT)
        )

    def _parse_time(self, time):
        time = time.strip().lower()
        if "am" in time or "pm" in time:
            return datetime.datetime.strptime(time, _AM_PM_TIMESTAMP_FORMAT)
        else:
            return datetime.datetime.strptime(time, _TIMESTAMP_FORMAT)



"""This module defines a command line tool for executing SQL queries on downloaded versions of
the CRDS catalog database.  It is a wrapper around a sqlite3 database download service and the
sqlite3 command line program which can be used to query  it.
"""
import sys
import os
import os.path
import sqlite3

from crds.core import log, cmdline, config
from crds.client import api

# ===================================================================

class CrdsSqlQueryScript(cmdline.Script):
    """Command line script for querying a downloaded copy of the CRDS catalog using sqlite3."""

    description = """crds.misc.sql is a thin wrapper around sqlite3 operating on a downloaded copy of
    the crds catalog or context history.   It is capable of doing basic select queries.    Fully general
    queries can be run under the external sqlite3 program operating on the dowloaded CRDS catalog
    sqlite3 database file.
"""

    epilog = """
* Dumping the table names supported can be done as follows:

crds sql --list-tables
crds_hst_catalog
crds_hst_context_history

* Dumping the columns of a table can be done as follows:

crds sql --list-columns
id, name, blob, state, blacklisted, rejected, observatory, instrument, filekind, type, derived_from, sha1sum, delivery_date, activation_date, useafter_date, change_level, pedigree, reference_file_type, size, uploaded_as, creator_name, deliverer_user, deliverer_email, description, catalog_link, replaced_by_filename, comment, aperture

* Executing a SQL query can be done as follows:

crds sql "select name, state, pedigree from crds_hst_catalog"
('h230851po_pfl.fits', 'operational', 'GROUND')
('h230851qo_pfl.fits', 'operational', 'GROUND')
('h230851so_pfl.fits', 'operational', 'GROUND')
('h2310096o_pfl.fits', 'operational', 'GROUND')
('h241320gn_ilm.fits', 'operational', 'DUMMY 04/02/1997')
('h241320in_ilm.fits', 'operational', 'DUMMY 04/02/1997')
('h241320jn_ilm.fits', 'operational', 'DUMMY 04/02/1997')
('h241320ln_ilm.fits', 'operational', 'DUMMY 04/02/1997')
...
"""

    def add_args(self):
        super(CrdsSqlQueryScript, self).add_args()
        self.add_argument("sql_query", nargs="*", help="sqlite3 query to run on cached database.")
        self.add_argument("-u", "--update-db", action="store_true",
                          help="Updates local cached copy of CRDS catalog and context history from CRDS server.")
        self.add_argument("-t", "--list-tables", action="store_true",
                          help="Print out the table names contained in the database")
        self.add_argument("-c", "--list-columns", metavar="TABLE", nargs="+",
                          help="Print out the column names contained in the specified table of the database")
        self.add_argument("-p", "--list-database-path", action="store_true",
                          help="Print out the path of the local sqlite3 database file,  useable by sqlite3.")

    def main(self):
        """Process command line parameters in to a context and list of
        reference files.   Print out the match tuples within the context
        which contain the reference files.
        """
        if self.args.update_db or not os.path.exists(self.sqlite_db_path):
            self.fetch_sqlite_db()

        if self.args.list_tables:
            self.list_tables()

        if self.args.list_columns:
            self.list_columns(self.args.list_columns[0])

        if self.args.list_database_path:
            print(self.sqlite_db_path)

        if self.args.sql_query:
            self.run_query(" ".join(self.args.sql_query))

        return log.errors()

    @property
    def sqlite_db_path(self):
        """Return the path to the sqlite3 database file."""
        return config.get_sqlite3_db_path(self.observatory)

    def fetch_sqlite_db(self):
        """Download a SQLite version of the CRDS catalog from the server."""
        log.info("Downloading CRDS catalog database file.")
        with log.warn_on_exception("Failed updating local CRDS database"):
            db_path = api.get_sqlite_db(self.observatory)
            log.info("Sqlite3 database file downloaded to:", db_path)

    def list_tables(self):
        """Print out the names of the database tables,  one per line."""
        for table in self.get_tables():
            print(table)

    def list_columns(self, table):
        """Print out an ordered comma separated list of database table column names"""
        print(", ".join(self.get_columns(table)))

    def get_tables(self):
        """Return the list of database table names."""
        connection = sqlite3.connect(self.sqlite_db_path)
        cursor = connection.cursor()
        query = 'select name from sqlite_master where type=\'table\''
        log.verbose("querying:", repr(query))
        tables = [row[0] for row in cursor.execute(query)]
        connection.close()
        return tables

    def get_columns(self, table):
        """Return the list of database column names for `table`."""
        connection = sqlite3.connect(self.sqlite_db_path)
        connection.row_factory = sqlite3.Row
        cursor = connection.cursor()
        query = "select * from {};".format(table)
        log.verbose("querying:", repr(query))
        row = cursor.execute(query).fetchone()
        columns = row.keys()
        connection.close()
        return columns

    def run_query(self, query):
        """Run the string `query` on the downloaded CRDS sqlite database."""
        connection = sqlite3.connect(self.sqlite_db_path)
        cursor = connection.cursor()
        log.verbose("querying:", repr(query))
        for row in cursor.execute(query):
            print(self.format_row(row))
        connection.commit()
        connection.close()

    def format_row(self, row):
        """Return a query row formatted for output."""
        row = squash_unicode(row)
        return row

# ===================================================================

def squash_unicode(row):
    """Convert unicode strings in row to ordinary strings."""
    row2 = []
    for field in row:
        if isinstance(field, str):
            row2.append(str(field))
        else:
            row2.append(field)
    return tuple(row2)

# ===================================================================

if __name__ == "__main__":
    sys.exit(CrdsSqlQueryScript()())



"""This module contains doctests that exercise the crds.sync module used to update and manage
a CRDS cache of rules and references.
"""
import os

import crds
from crds.core import config, rmap
from crds.sync import SyncScript
from crds.tests import test_config

# ==================================================================================

class TestSync(test_config.CRDSTestCase):

    script_class = SyncScript

    def setUp(self):
        super(TestSync, self).setUp()
        os.environ["CRDS_PATH"] = self.temp_dir
        os.environ["CRDS_REF_SUBDIR_MODE"] = "flat"

    def test_sync_contexts(self):
        self.run_script("crds.sync --contexts hst_cos.imap")
        for name in crds.get_cached_mapping("hst_cos.imap").mapping_names():
            self.assert_crds_exists(name)
        self.run_script("crds.sync --contexts hst_cos_deadtab.rmap --fetch-references")
        for name in crds.get_cached_mapping("hst_cos_deadtab.rmap").reference_names():
            self.assert_crds_exists(name)
            with open(config.locate_file(name, "hst"), "w+") as handle:
                handle.write("foo")
        self.run_script("crds.sync --contexts hst_cos_deadtab.rmap --fetch-references --check-files", 2)
        self.run_script("crds.sync --contexts hst_cos_deadtab.rmap --fetch-references --check-files --repair-files", 2)
        self.run_script("crds.sync --contexts hst_cos_deadtab.rmap --fetch-references --check-files --repair-files")
        self.run_script("crds.sync --contexts hst_cos_deadtab.rmap --fetch-references --check-files --repair-files --check-sha1sum")

    def test_sync_explicit_files(self):
        self.assert_crds_not_exists("hst_cos_deadtab.rmap")
        self.run_script("crds.sync --files hst_cos_deadtab.rmap --check-files --repair-files --check-sha1sum")
        crds.get_cached_mapping("hst_cos_deadtab.rmap")

    def test_sync_readonly_cache(self):
        super(TestSync, self).setUp()
        # self.tearDown()   # switch to default test environment (currently) from hst-crds-dev.csh
        self.run_script("crds.sync --contexts hst_cos_deadtab.rmap")  # sync to establish config area for sure
        self.run_script("crds.sync --contexts hst_cos_deadtab.rmap --fetch-references --readonly-cache")

    def test_sync_organize_flat(self):
        self.assert_crds_not_exists("hst_cos_deadtab.rmap")
        self.assert_crds_not_exists("s7g1700gl_dead.fits")
        self.assert_crds_not_exists("s7g1700ql_dead.fits")
        self.run_script("crds.sync --contexts hst_cos_deadtab.rmap --fetch-references --organize=flat")
        self.assert_crds_exists("hst_cos_deadtab.rmap")
        self.assert_crds_exists("s7g1700gl_dead.fits")
        self.assert_crds_exists("s7g1700ql_dead.fits")

    def test_sync_organize_instr(self):
        self.run_script("crds.sync --contexts hst_cos_deadtab.rmap --fetch-references --organize=instrument --organize-delete-junk")
        self.assert_crds_exists("hst_cos_deadtab.rmap")
        self.assert_crds_exists("s7g1700gl_dead.fits")
        self.assert_crds_exists("s7g1700ql_dead.fits")
        self.run_script("crds.sync --contexts hst_cos_deadtab.rmap --fetch-references --organize=flat --organize-delete-junk")

    def test_sync_fetch_sqlite3_db(self):
        self.run_script("crds.sync --fetch-sqlite-db")

    def test_sync_dataset_ids(self):
        old_state = test_config.setup()
        self.run_script("crds.sync --contexts hst.pmap --dataset-ids LA9K03CBQ:LA9K03CBQ --fetch-references")
        test_config.cleanup(old_state)

    def test_purge_mappings(self):
        self.run_script("crds.sync --contexts hst_cos_deadtab.rmap --fetch-references")
        self.run_script("crds.sync --organize=flat")
        r = crds.get_cached_mapping("hst_cos_deadtab.rmap")
        self.assertEqual(r.reference_names(), ["s7g1700gl_dead.fits", "s7g1700ql_dead.fits"])
        self.assertEqual(rmap.list_references("*", "hst"), ["s7g1700gl_dead.fits", "s7g1700ql_dead.fits"])
        self.assert_crds_exists("s7g1700gl_dead.fits")
        self.assert_crds_exists("s7g1700ql_dead.fits")
        self.run_script("crds.sync --contexts hst_acs_imphttab.rmap --fetch-references --purge-mappings --purge-references")
        self.assertEqual(rmap.list_references("*", "hst"), ['w3m1716tj_imp.fits', 'w3m17170j_imp.fits', 'w3m17171j_imp.fits'])
        self.assertEqual(rmap.list_mappings("*", "hst"), ['hst_acs_imphttab.rmap'])

# ==================================================================================


def tst():
    """Run module tests,  for now just doctests only."""
    import unittest
    suite = unittest.TestLoader().loadTestsFromTestCase(TestSync)
    unittest.TextTestRunner().run(suite)

if __name__ == "__main__":
    print(tst())



#! /usr/bin/env python
#-*-python-*-

"""This script is used to "check out" a version of synphot files corresponding
to a CRDS context with an organization compatible with the classic (Py)Synphot
directory structure.  This largely isolates CRDS from the CDBS/IRAF-style
synphot directory structure and keeps rmaps in a simple form which does not
include IRAF-style env vars as part of filenames.
"""

import os
import os.path
import sys
import shutil
import warnings

# =============================================================================

# from pysynphot import locations

SYNPHOT_IGNORE = [
    "Extinction files not found",
    "No graph or component tables found",
    "No thermal tables found"
    ]

# =============================================================================

import crds
from crds.core import pysh, log, utils, cmdline, config
from crds import sync
from crds import client
from crds.io import tables

# =============================================================================

class GetSynphotScript(cmdline.Script):
    """Command line script for modifying .rmap files."""

    description = """
This script downloads (Py)Synphot files refered to by <context>
from the CRDS server to directory <synphot_dir>.
    """

    epilog = """
get_synphot works by first creating a CRDS cache under the specified
<synphot_dir> and then hard linking files from that CRDS cache to their
classic CDBS directory equivalents in comp and mtab subdirectories.

With --keep-crds get_synphot will retain the CRDS cache structure used to
perform the initial download of files under <synphot_dir>/crds. This makes
additional calls to get_synphot incremental updates instead of full downloads.
The <synphot_dir>/crds directory exists in addition to the mtab and comp
directories.

Example get_synophot runs:

1. Basic run

  $ crds get_synphot  my_pysyn_dir  hst-edit

  Overwrites my_pysyn_dir in the current working directory, storing the (py)synphot
  reference files associated with with the latest file deliveries to HST.   This
  will force a full download and will remove existing synphot subdirectories and
  files whether they come from CRDS or not.

2. Get specific version of synphot files (hyothetical context hst_0999.pmap)

  $ crds get_synphot  my_pysyn_dir  hst_0999.pmap

3. Get synphot files,  keeping working copy of CRDS cache

  $ crds get_synphot  my_pysyn_dir  hst-edit --keep-crds

  This keeps a copy of the working CRDS cache under <my_pysyn_dir>/crds.  These files
  refer to the same storage as the normal synphot paths and do not require extra space.
  --keep-crds is largely used to minimize repeat downloads.

4. Get synphot files,  keeping old versions and user added synphot files

  $ crds get_synphot  my_pysyn_dir  hst-edit --keep-synphot

  This retains the existing synphot comp and mtab directories my_pysyn_dir.
"""
    def __init__(self, *args, **keys):
        super(GetSynphotScript, self).__init__(*args, **keys)
        self.pmap = None
        self.imap = None

    def add_args(self):
        """Add command line parameters specific to this script."""
        self.add_argument(
            "synphot_dir", type=str, help="Location of synphot directory.")
        self.add_argument(
            'context', type=cmdline.context_spec,
            help="CRDS context (.pmap) defining version of synphot references to fetch.")
        self.add_argument(
            "--keep-crds", action="store_true",
            help="Don't remove the 'crds' cache subdirectory of <synphot_dir> after downloading.")
        self.add_argument(
            "--keep-synphot", action="store_true",
            help="Don't remove <synphot_dir> subdirectories before linking requested files.")

    def determine_contexts(self):
        """Use the command line `context` parameter to help define things like
        observatory and default CRDS_SERVER_URL.
        """
        return [self.args.context]

    # -------------------------------------------------------------------------

    def main(self):
        """Perform the high level sequence of tasks needed to download and
        organize a version of pysynphot files under the specified directory.
        """
        self.deferred_init()

        # Blow away CRDS cache prior to syncing
        if not self.args.keep_crds:
            self.rmdir("crds")

        self.crds_download()

        syn_name_map = self.pysyn_cdbs_from_syn_tables()

        # Blow away comp and mtab to remove old versions of links
        if not self.args.keep_synphot:
            self.rmdir("comp")
            self.rmdir("mtab")

        self.cross_link_cdbs_paths(syn_name_map)

        # Blow away CRDS cache leaving only synphot organization.
        # in the final product.
        if not self.args.keep_crds:
            self.rmdir("crds")

        log.standard_status()
        return log.errors()

    # -------------------------------------------------------------------------

    def deferred_init(self):
        """Initialize CRDS and PySYNPHOT after module import and basic script
        object construction.
        """
        self.crds_cache = os.path.abspath(os.path.join(
            self.args.synphot_dir, "crds"))
        config.override_crds_paths(self.crds_cache)
        os.environ["PYSYN_CDBS"] = self.args.synphot_dir
        utils.clear_function_caches()

    def crds_download(self):
        """Populates the <synphot_dir>/crds file cache with all synphot files
        implied by self.args.context.   This creates an ordinary CRDS cache
        within <synphot_dir> which both sources the synphot files needed now
        and can optimize future downloads if desired.
        """
        resolved_context = self.resolve_context(self.args.context)
        self.dump_mappings([resolved_context])
        self.pmap = crds.get_cached_mapping(resolved_context)
        self.imap = self.pmap.get_imap("synphot")
        self.dump_files(resolved_context, self.imap.reference_names())

    # -------------------------------------------------------------------------

    def pysyn_cdbs_from_syn_tables(self):
        """Construct the mapping from synphot file basename to classic CDBS path for
        every file defined by the TMC and TMT tables.  Also include mtab paths
        for the master TMC,TMT,TMG tables themselves.

        This bridges the gap between basenames recorded by CRDS and synphot
        paths which are defined with environment variables encoded in the table
        filenames.

        Reads table files ending with _tmt.fits and _tmc.fits

        Returns   { syn_basename : full_pysn_path, ... }
        """
        syn_name_map = dict()
        syn_name_map.update(self.get_comptab_info("tmt"))
        syn_name_map.update(self.get_comptab_info("tmc"))
        syn_name_map.update(self.get_mtab_info("tmc"))
        syn_name_map.update(self.get_mtab_info("tmt"))
        syn_name_map.update(self.get_mtab_info("tmg"))
        return syn_name_map

    def get_comptab_info(self, synname):
        """Dump the FILENAME column of the component table implied by `synname`
        (e.g. "tmc" --> something_tmc.fits) and use pysynphot to interpret the
        embedded iraf$-style path env var into a normal filepath.  This is used
        to locate files within appropriate sub-directories of
        <synphot_dir>/comp.

        Return the mapping from a component file basename as defined in the CRDS
        rmap to the absolute path in a CDBS-style synphot file repo being created.

        Returns  { component_basename : abs_pysyn_path, ...}
        """
        for msg in SYNPHOT_IGNORE:
            warnings.filterwarnings("ignore",msg)
        from pysynphot import locations

        filekind = synname + "tab"
        rmap  = self.imap.get_rmap(filekind)

        references  = rmap.reference_names()
        assert len(references) == 1, \
            "More than one '%s' reference name mentioned in '%s'." % \
            (synname, rmap.name)
        tab_name = references[0]

        # rmap object locate() not module function.
        tab_path = rmap.locate_file(tab_name)

        # CRDS abstract table object nominally from HDU 1
        table = tables.tables(tab_path)[0]

        fileinfo = {}
        for syn_name in table.columns["FILENAME"]:

            iraf_path, basename = syn_name.split("$")
            name = basename.split("[")[0]  # remove parameterization

            dollar_syn_name = syn_name.split("[")[0]

            # Use pysynphot to interpret iraf_path
            cdbs_filepath = os.path.abspath(
                locations.irafconvert(dollar_syn_name))

            fileinfo[name] = cdbs_filepath

        return fileinfo

    def get_mtab_info(self, tab):
        """Return the mapping from basename to the full pysyn CDBS-style absolute
        filepath for any master TMC,TMG, or TMT table file referred to under
        the spoecified context.  There can be only one of each kind of file.

        tab       string      nominally one of "tmc","tmt","tmg"

        Returns   { table_basename(tab) : pysyn_mtab_filepath }
        """
        rmap = self.imap.get_rmap(tab + "tab")
        references = rmap.reference_names()
        assert len(references) == 1, "More than one " + repr(tab.upper()) + " file."
        filename = references[0]
        fullpath = os.path.abspath(os.path.join(
            self.args.synphot_dir, "mtab", filename))
        return { filename : fullpath }

    # -------------------------------------------------------------------------

    def cross_link_cdbs_paths(self, syn_name_map):
        """Hard link files from the downloaded CRDS cache structure to the
        classic CDBS-style directory locations specified by `syn_name_map`.

        On OS-X and Linux this creates files which are referenced from two
        different paths on the file system.   Deleting one path or the other
        leaves behind a normal file referenced from only one location.

        syn_name_map    dict     { syn_basename : pysyn_filepath, ... }

        returns   None
        """
        for reference in syn_name_map:
            with log.error_on_exception(
                    "Failed linking", repr(reference), "to CDBS directory."):
                crds_filepath = os.path.abspath(self.imap.locate_file(reference))
                cdbs_filepath = syn_name_map[reference]
                utils.ensure_dir_exists(cdbs_filepath)
                if not os.path.exists(cdbs_filepath):
                    os.link(crds_filepath, cdbs_filepath)
                    log.verbose("Linked", cdbs_filepath, "-->", crds_filepath)
                else:
                    log.verbose("Skipped existing", cdbs_filepath)

    def rmdir(self, subdir):
        """If it exists, remove `subdir` of the <synphot_dir> specified.

        subdir    string    subdirectory of <synphot_dir> to remove.
        """
        path = os.path.join(self.args.synphot_dir, subdir)
        log.verbose("rmdir: ", repr(path))
        with log.verbose_warning_on_exception("Failed removing", repr(path)):
            shutil.rmtree(path)

# -----------------------------------------------------------------------------

if __name__ == "__main__":
    sys.exit(GetSynphotScript()())
